# Copyright (C) 2019 Cyrus Cheng (Imperial College London)
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by David A. Ham (david.ham@imperial.ac.uk), 2019

import numbers
import sympy
from sympy import symbols, legendre, Array, diff, lambdify
import numpy as np
from FIAT.finite_element import FiniteElement
from FIAT.lagrange import Lagrange
from FIAT.dual_set import make_entity_closure_ids
from FIAT.polynomial_set import mis
from FIAT.reference_element import (compute_unflattening_map,
                                    flatten_reference_cube)
from FIAT.reference_element import make_lattice

from FIAT.pointwise_dual import compute_pointwise_dual

x, y, z = symbols('x y z')
variables = (x, y, z)
leg = legendre


def tr(n):
    if n <= 1:
        return 0
    else:
        return int((n-3)*(n-2)/2)


def _replace_numbers_with_symbols(polynomials):
    # Replace numbers with symbols to work around issue with numpy>=1.24.1;
    # see https://github.com/firedrakeproject/fiat/pull/32.
    extra_vars = {}  # map from numbers to symbols
    polynomials_list = []
    for poly in polynomials.tolist():
        if isinstance(poly, numbers.Real):
            if poly not in extra_vars:
                extra_vars[poly] = symbols('num_' + str(len(extra_vars)))
            polynomials_list.append(extra_vars[poly])
        elif isinstance(poly, sympy.core.Expr):
            polynomials_list.append(poly)
        else:
            raise TypeError(f"Unexpected type: {type(poly)}")
    polynomials = Array(polynomials_list)
    return polynomials, extra_vars


class Serendipity(FiniteElement):

    def __new__(cls, ref_el, degree):
        dim = ref_el.get_spatial_dimension()
        if dim == 1:
            return Lagrange(ref_el, degree)
        elif dim == 0:
            raise IndexError("reference element cannot be dimension 0")
        else:
            self = super().__new__(cls)
            return self

    def __init__(self, ref_el, degree):

        flat_el = flatten_reference_cube(ref_el)
        dim = flat_el.get_spatial_dimension()
        flat_topology = flat_el.get_topology()

        verts = flat_el.get_vertices()

        dx = ((verts[-1][0] - x)/(verts[-1][0] - verts[0][0]), (x - verts[0][0])/(verts[-1][0] - verts[0][0]))
        dy = ((verts[-1][1] - y)/(verts[-1][1] - verts[0][1]), (y - verts[0][1])/(verts[-1][1] - verts[0][1]))
        x_mid = 2*x-(verts[-1][0] + verts[0][0])
        y_mid = 2*y-(verts[-1][1] + verts[0][1])
        try:
            dz = ((verts[-1][2] - z)/(verts[-1][2] - verts[0][2]), (z - verts[0][2])/(verts[-1][2] - verts[0][2]))
            z_mid = 2*z-(verts[-1][2] + verts[0][2])
        except IndexError:
            dz = None
            z_mid = None

        VL = v_lambda_0(dim, dx, dy, dz)
        EL = []
        FL = []
        IL = []
        s_list = []
        entity_ids = {}
        cur = 0

        for top_dim, entities in flat_topology.items():
            entity_ids[top_dim] = {}
            for entity in entities:
                entity_ids[top_dim][entity] = []

        for j in sorted(flat_topology[0]):
            entity_ids[0][j] = [cur]
            cur = cur + 1

        EL += e_lambda_0(degree, dim, dx, dy, dz, x_mid, y_mid, z_mid)

        for j in sorted(flat_topology[1]):
            entity_ids[1][j] = list(range(cur, cur + degree - 1))
            cur = cur + degree - 1

        FL += f_lambda_0(degree, dim, dx, dy, dz, x_mid, y_mid, z_mid)

        for j in sorted(flat_topology[2]):
            entity_ids[2][j] = list(range(cur, cur + tr(degree)))
            cur = cur + tr(degree)

        if dim == 3:
            IL += i_lambda_0(degree, dx, dy, dz, x_mid, y_mid, z_mid)

            entity_ids[3] = {}
            entity_ids[3][0] = list(range(cur, cur + len(IL)))
            cur = cur + len(IL)

        s_list = VL + EL + FL + IL
        assert len(s_list) == cur
        formdegree = 0

        super().__init__(ref_el=ref_el, dual=None, order=degree, formdegree=formdegree)

        self.basis = {(0,)*dim: Array(s_list)}
        polynomials, extra_vars = _replace_numbers_with_symbols(Array(s_list))
        self.basis_callable = {(0,)*dim: [lambdify(variables[:dim], polynomials,
                                                   modules="numpy", dummify=True),
                                          extra_vars]}
        topology = ref_el.get_topology()
        unflattening_map = compute_unflattening_map(topology)
        unflattened_entity_ids = {}
        unflattened_entity_closure_ids = {}

        entity_closure_ids = make_entity_closure_ids(flat_el, entity_ids)

        for dim, entities in sorted(topology.items()):
            unflattened_entity_ids[dim] = {}
            unflattened_entity_closure_ids[dim] = {}
        for dim, entities in sorted(flat_topology.items()):
            for entity in entities:
                unflat_dim, unflat_entity = unflattening_map[(dim, entity)]
                unflattened_entity_ids[unflat_dim][unflat_entity] = entity_ids[dim][entity]
                unflattened_entity_closure_ids[unflat_dim][unflat_entity] = entity_closure_ids[dim][entity]
        self.entity_ids = unflattened_entity_ids
        self.entity_closure_ids = unflattened_entity_closure_ids
        self._degree = degree
        self.flat_el = flat_el

        self.dual = compute_pointwise_dual(self, unisolvent_pts(ref_el, degree))

    def degree(self):
        return self._degree + 1

    def get_nodal_basis(self):
        raise NotImplementedError("get_nodal_basis not implemented for serendipity")

    def get_dual_set(self):
        raise NotImplementedError("get_dual_set is not implemented for serendipity")

    def get_coeffs(self):
        raise NotImplementedError("get_coeffs not implemented for serendipity")

    def tabulate(self, order, points, entity=None):

        if entity is None:
            entity = (self.ref_el.get_dimension(), 0)

        entity_dim, entity_id = entity
        transform = self.ref_el.get_entity_transform(entity_dim, entity_id)
        points = transform(points)

        phivals = {}
        dim = self.flat_el.get_spatial_dimension()
        if dim <= 1:
            raise NotImplementedError('no tabulate method for serendipity elements of dimension 1 or less.')
        if dim >= 4:
            raise NotImplementedError('tabulate does not support higher dimensions than 3.')
        npoints, pointdim = points.shape
        for o in range(order + 1):
            alphas = mis(dim, o)
            for alpha in alphas:
                try:
                    callable, extra_vars = self.basis_callable[alpha]
                except KeyError:
                    polynomials = diff(self.basis[(0,)*dim], *zip(variables, alpha))
                    polynomials, extra_vars = _replace_numbers_with_symbols(polynomials)
                    callable = lambdify(variables[:dim] + tuple(extra_vars.values()), polynomials, modules="numpy", dummify=True)
                    self.basis[alpha] = polynomials
                    self.basis_callable[alpha] = [callable, extra_vars]
                # Can no longer make a numpy array from objects of inhomogeneous shape
                # (unless we specify `dtype==object`);
                # see https://github.com/firedrakeproject/fiat/pull/32.
                #
                # Casting `key`s to float() is needed, otherwise we somehow get the following error:
                #
                # E           TypeError: unsupported type for persistent hash keying: <class 'complex'>
                #
                # ../../lib/python3.8/site-packages/pytools/persistent_dict.py:243: TypeError
                #
                # `key`s have been checked to be numbers.Real.
                extra_arrays = [np.ones((npoints, ), dtype=points.dtype) * float(key) for key in extra_vars]
                phivals[alpha] = callable(*([points[:, i] for i in range(pointdim)] + extra_arrays))
        return phivals

    def entity_dofs(self):
        """Return the map of topological entities to degrees of
        freedom for the finite element."""
        return self.entity_ids

    def entity_closure_dofs(self):
        """Return the map of topological entities to degrees of
        freedom on the closure of those entities for the finite element."""
        return self.entity_closure_ids

    def value_shape(self):
        return ()

    def dmats(self):
        raise NotImplementedError

    def get_num_members(self, arg):
        raise NotImplementedError

    def space_dimension(self):
        return len(self.basis[(0,)*self.flat_el.get_spatial_dimension()])


def v_lambda_0(dim, dx, dy, dz):

    if dim == 2:
        VL = [a*b for a in dx for b in dy]
    else:
        VL = [a*b*c for a in dx for b in dy for c in dz]

    return VL


def e_lambda_0(i, dim, dx, dy, dz, x_mid, y_mid, z_mid):

    if dim == 2:
        EL = tuple([-leg(j, y_mid) * dy[0] * dy[1] * a for a in dx for j in range(i-1)] +
                   [-leg(j, x_mid) * dx[0] * dx[1] * b for b in dy for j in range(i-1)])
    else:
        EL = tuple([-leg(j, z_mid) * dz[0] * dz[1] * a * b for b in dx for a in dy for j in range(i-1)] +
                   [-leg(j, y_mid) * dy[0] * dy[1] * a * c for a in dx for c in dz for j in range(i-1)] +
                   [-leg(j, x_mid) * dx[0] * dx[1] * b * c for c in dy for b in dz for j in range(i-1)])

    return EL


def f_lambda_0(i, dim, dx, dy, dz, x_mid, y_mid, z_mid):

    if dim == 2:
        FL = tuple([leg(j, x_mid) * leg(k-4-j, y_mid) * dx[0] * dx[1] * dy[0] * dy[1]
                    for k in range(4, i + 1) for j in range(k-3)])
    else:
        FL = tuple([leg(j, y_mid) * leg(k-4-j, z_mid) * dy[0] * dy[1] * dz[0] * dz[1] * a
                    for a in dx for k in range(4, i + 1) for j in range(k-3)] +
                   [leg(j, z_mid) * leg(k-4-j, x_mid) * dx[0] * dx[1] * dz[0] * dz[1] * b
                    for b in dy for k in range(4, i + 1) for j in range(k-3)] +
                   [leg(j, x_mid) * leg(k-4-j, y_mid) * dx[0] * dx[1] * dy[0] * dy[1] * c
                    for c in dz for k in range(4, i + 1) for j in range(k-3)])

    return FL


def i_lambda_0(i, dx, dy, dz, x_mid, y_mid, z_mid):

    IL = tuple([-leg(l-6-j, x_mid) * leg(j-k, y_mid) * leg(k, z_mid) *
                dx[0] * dx[1] * dy[0] * dy[1] * dz[0] * dz[1]
                for l in range(6, i + 1) for j in range(l-5) for k in range(j+1)])

    return IL


def unisolvent_pts(K, deg):
    flat_el = flatten_reference_cube(K)
    dim = flat_el.get_spatial_dimension()
    if dim == 2:
        return unisolvent_pts_quad(flat_el, deg)
    elif dim == 3:
        return unisolvent_pts_hex(flat_el, deg)
    else:
        raise ValueError("Serendipity only defined for quads and hexes")


def unisolvent_pts_quad(K, deg):
    """Gives a set of unisolvent points for the quad serendipity space of order deg.
    The S element is not dual to these nodes, but a dual basis can be constructed from them."""
    L = K.construct_subelement(1)
    vs = np.asarray(K.vertices)
    pts = [pt for pt in K.vertices]
    Lpts = make_lattice(L.vertices, deg, 1)
    for e in K.topology[1]:
        Fmap = K.get_entity_transform(1, e)
        epts = [tuple(Fmap(pt)) for pt in Lpts]
        pts.extend(epts)
    if deg > 3:
        dx0 = (vs[1, :] - vs[0, :]) / (deg-2)
        dx1 = (vs[2, :] - vs[0, :]) / (deg-2)

        internal_nodes = [tuple(vs[0, :] + dx0 * i + dx1 * j)
                          for i in range(1, deg-2)
                          for j in range(1, deg-1-i)]
        pts.extend(internal_nodes)

    return pts


def unisolvent_pts_hex(K, deg):
    """Gives a set of unisolvent points for the hex serendipity space of order deg.
    The S element is not dual to these nodes, but a dual basis can be constructed from them."""
    L = K.construct_subelement(1)
    F = K.construct_subelement(2)
    vs = np.asarray(K.vertices)
    pts = [pt for pt in K.vertices]
    Lpts = make_lattice(L.vertices, deg, 1)
    for e in K.topology[1]:
        Fmap = K.get_entity_transform(1, e)
        epts = [tuple(Fmap(pt)) for pt in Lpts]
        pts.extend(epts)
    if deg > 3:
        fvs = np.asarray(F.vertices)
        # Planar points to map to each face
        dx0 = (fvs[1, :] - fvs[0, :]) / (deg-2)
        dx1 = (fvs[2, :] - fvs[0, :]) / (deg-2)

        Fpts = [tuple(fvs[0, :] + dx0 * i + dx1 * j)
                for i in range(1, deg-2)
                for j in range(1, deg-1-i)]
        for f in K.topology[2]:
            Fmap = K.get_entity_transform(2, f)
            pts.extend([tuple(Fmap(pt)) for pt in Fpts])
    if deg > 5:
        dx0 = np.asarray([1., 0, 0]) / (deg-4)
        dx1 = np.asarray([0, 1., 0]) / (deg-4)
        dx2 = np.asarray([0, 0, 1.]) / (deg-4)

        Ipts = [tuple(vs[0, :] + dx0 * i + dx1 * j + dx2 * k)
                for i in range(1, deg-4)
                for j in range(1, deg-3-i)
                for k in range(1, deg-2-i-j)]
        pts.extend(Ipts)

    return pts



"""
This file contains all the necessary functions to accurately count the
total number of floating point operations for a given script.
"""

import gem.gem as gem
import gem.impero as imp
from functools import singledispatch
import numpy
import math


@singledispatch
def statement(tree, temporaries):
    raise NotImplementedError


@statement.register(imp.Block)
def statement_block(tree, temporaries):
    flops = sum(statement(child, temporaries) for child in tree.children)
    return flops


@statement.register(imp.For)
def statement_for(tree, temporaries):
    extent = tree.index.extent
    assert extent is not None
    child, = tree.children
    flops = statement(child, temporaries)
    return flops * extent


@statement.register(imp.Initialise)
def statement_initialise(tree, temporaries):
    return 0


@statement.register(imp.Accumulate)
def statement_accumulate(tree, temporaries):
    flops = expression_flops(tree.indexsum.children[0], temporaries)
    return flops + 1


@statement.register(imp.Return)
def statement_return(tree, temporaries):
    flops = expression_flops(tree.expression, temporaries)
    return flops + 1


@statement.register(imp.ReturnAccumulate)
def statement_returnaccumulate(tree, temporaries):
    flops = expression_flops(tree.indexsum.children[0], temporaries)
    return flops + 1


@statement.register(imp.Evaluate)
def statement_evaluate(tree, temporaries):
    flops = expression_flops(tree.expression, temporaries, top=True)
    return flops


@singledispatch
def flops(expr, temporaries):
    raise NotImplementedError(f"Don't know how to count flops of {type(expr)}")


@flops.register(gem.Failure)
def flops_failure(expr, temporaries):
    raise ValueError("Not expecting a Failure node")


@flops.register(gem.Variable)
@flops.register(gem.Identity)
@flops.register(gem.Delta)
@flops.register(gem.Zero)
@flops.register(gem.Literal)
@flops.register(gem.Index)
@flops.register(gem.VariableIndex)
def flops_zero(expr, temporaries):
    # Initial set up of these Gem nodes are of 0 floating point operations.
    return 0


@flops.register(gem.LogicalNot)
@flops.register(gem.LogicalAnd)
@flops.register(gem.LogicalOr)
@flops.register(gem.ListTensor)
def flops_zeroplus(expr, temporaries):
    # These nodes contribute 0 floating point operations, but their children may not.
    return 0 + sum(expression_flops(child, temporaries)
                   for child in expr.children)


@flops.register(gem.Product)
def flops_product(expr, temporaries):
    # Multiplication by -1 is not a flop.
    a, b = expr.children
    if isinstance(a, gem.Literal) and a.value == -1:
        return expression_flops(b, temporaries)
    elif isinstance(b, gem.Literal) and b.value == -1:
        return expression_flops(a, temporaries)
    else:
        return 1 + sum(expression_flops(child, temporaries)
                       for child in expr.children)


@flops.register(gem.Sum)
@flops.register(gem.Division)
@flops.register(gem.Comparison)
@flops.register(gem.MathFunction)
@flops.register(gem.MinValue)
@flops.register(gem.MaxValue)
def flops_oneplus(expr, temporaries):
    return 1 + sum(expression_flops(child, temporaries)
                   for child in expr.children)


@flops.register(gem.Power)
def flops_power(expr, temporaries):
    base, exponent = expr.children
    base_flops = expression_flops(base, temporaries)
    if isinstance(exponent, gem.Literal):
        exponent = exponent.value
        if exponent > 0 and exponent == math.floor(exponent):
            return base_flops + int(math.ceil(math.log2(exponent)))
        else:
            return base_flops + 5  # heuristic
    else:
        return base_flops + 5   # heuristic


@flops.register(gem.Conditional)
def flops_conditional(expr, temporaries):
    condition, then, else_ = (expression_flops(child, temporaries)
                              for child in expr.children)
    return condition + max(then, else_)


@flops.register(gem.Indexed)
@flops.register(gem.FlexiblyIndexed)
def flops_indexed(expr, temporaries):
    aggregate = sum(expression_flops(child, temporaries)
                    for child in expr.children)
    # Average flops per entry
    return aggregate / numpy.prod(expr.children[0].shape, dtype=int)


@flops.register(gem.IndexSum)
def flops_indexsum(expr, temporaries):
    raise ValueError("Not expecting IndexSum")


@flops.register(gem.Inverse)
def flops_inverse(expr, temporaries):
    n, _ = expr.shape
    # 2n^3 + child flop count
    return 2*n**3 + sum(expression_flops(child, temporaries)
                        for child in expr.children)


@flops.register(gem.Solve)
def flops_solve(expr, temporaries):
    n, m = expr.shape
    # 2mn + inversion cost of A + children flop count
    return 2*n*m + 2*n**3 + sum(expression_flops(child, temporaries)
                                for child in expr.children)


@flops.register(gem.ComponentTensor)
def flops_componenttensor(expr, temporaries):
    raise ValueError("Not expecting ComponentTensor")


def expression_flops(expression, temporaries, top=False):
    """An approximation to flops required for each expression.

    :arg expression: GEM expression.
    :arg temporaries: Expressions that are assigned to temporaries
    :arg top: are we at the root?
    :returns: flop count for the expression
    """
    if not top and expression in temporaries:
        return 0
    else:
        return flops(expression, temporaries)


def count_flops(impero_c):
    """An approximation to flops required for a scheduled impero_c tree.

    :arg impero_c: a :class:`~.Impero_C` object.
    :returns: approximate flop count for the tree.
    """
    try:
        return statement(impero_c.tree, set(impero_c.temporaries))
    except (ValueError, NotImplementedError):
        return 0



import FIAT
from gem import ListTensor

from finat.fiat_elements import ScalarFiatElement
from finat.physically_mapped import Citations, identity, PhysicallyMappedElement


class Hermite(PhysicallyMappedElement, ScalarFiatElement):
    def __init__(self, cell, degree=3):
        if Citations is not None:
            Citations().register("Ciarlet1972")
        super().__init__(FIAT.CubicHermite(cell))

    def basis_transformation(self, coordinate_mapping):
        Js = [coordinate_mapping.jacobian_at(vertex)
              for vertex in self.cell.get_vertices()]

        h = coordinate_mapping.cell_size()

        d = self.cell.get_dimension()
        M = identity(self.space_dimension())

        cur = 0
        for i in range(d+1):
            cur += 1  # skip the vertex
            J = Js[i]
            for j in range(d):
                for k in range(d):
                    M[cur+j, cur+k] = J[j, k] / h[i]
            cur += d

        return ListTensor(M)



import FIAT

from finat.physically_mapped import Citations
from finat.piola_mapped import PiolaBubbleElement


class ChristiansenHu(PiolaBubbleElement):
    def __init__(self, cell, degree=1):
        if Citations is not None:
            Citations().register("ChristiansenHu2019")
        super().__init__(FIAT.ChristiansenHu(cell, degree))



# Copyright (C) 2015 Imperial College London and others.
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Written by Robert C. Kirby (robert_kirby@baylor.edu), 2020


from FIAT import finite_element, polynomial_set, dual_set, functional, quadrature
from FIAT.reference_element import LINE


class GaussRadauDualSet(dual_set.DualSet):
    """The dual basis for 1D discontinuous elements with nodes at the
    Gauss-Radau points."""
    def __init__(self, ref_el, degree, right=True):
        # Do DG connectivity because it's bonkers to do one-sided assembly even
        # though we have an endpoint in the point set!
        entity_ids = {0: {0: [], 1: []},
                      1: {0: list(range(0, degree+1))}}
        lr = quadrature.RadauQuadratureLineRule(ref_el, degree+1, right)
        nodes = [functional.PointEvaluation(ref_el, x) for x in lr.pts]

        super().__init__(nodes, ref_el, entity_ids)


class GaussRadau(finite_element.CiarletElement):
    """1D discontinuous element with nodes at the Gauss-Radau points."""
    def __init__(self, ref_el, degree):
        if ref_el.shape != LINE:
            raise ValueError("Gauss-Radau elements are only defined in one dimension.")
        poly_set = polynomial_set.ONPolynomialSet(ref_el, degree)
        dual = GaussRadauDualSet(ref_el, degree)
        formdegree = ref_el.get_spatial_dimension()  # n-form
        super().__init__(poly_set, dual, degree, formdegree)



# -*- coding: utf-8 -*-
"""Implementation of the Arnold-Winther finite elements."""

# Copyright (C) 2020 by Robert C. Kirby (Baylor University)
# Modified by Francis Aznaran (Oxford University)
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later


from FIAT import finite_element, dual_set, polynomial_set
from FIAT.reference_element import TRIANGLE
from FIAT.quadrature_schemes import create_quadrature
from FIAT.functional import (ComponentPointEvaluation,
                             TensorBidirectionalIntegralMoment,
                             IntegralMomentOfTensorDivergence,
                             IntegralLegendreNormalNormalMoment,
                             IntegralLegendreNormalTangentialMoment)

import numpy


class ArnoldWintherNCDual(dual_set.DualSet):
    def __init__(self, ref_el, degree=2):
        if degree != 2:
            raise ValueError("Nonconforming Arnold-Winther elements are only defined for degree 2.")
        top = ref_el.get_topology()
        sd = ref_el.get_spatial_dimension()
        entity_ids = {dim: {entity: [] for entity in sorted(top[dim])} for dim in sorted(top)}
        nodes = []

        # no vertex dofs
        # proper edge dofs now (not the contraints)
        # edge dofs: bidirectional nn and nt moments against P1.
        qdegree = degree + 2
        for entity in sorted(top[1]):
            cur = len(nodes)
            for order in range(2):
                nodes.append(IntegralLegendreNormalNormalMoment(ref_el, entity, order, qdegree))
                nodes.append(IntegralLegendreNormalTangentialMoment(ref_el, entity, order, qdegree))
            entity_ids[1][entity].extend(range(cur, len(nodes)))

        # internal dofs: constant moments of three unique components
        cur = len(nodes)
        n = list(map(ref_el.compute_scaled_normal, sorted(top[sd-1])))
        Q = create_quadrature(ref_el, degree)
        phi = numpy.full(Q.get_weights().shape, 1/ref_el.volume())
        nodes.extend(TensorBidirectionalIntegralMoment(ref_el, n[i+1], n[j+1], Q, phi)
                     for i in range(sd) for j in range(i, sd))
        entity_ids[2][0].extend(range(cur, len(nodes)))

        # put the constraint dofs last.
        for entity in sorted(top[1]):
            cur = len(nodes)
            nodes.append(IntegralLegendreNormalNormalMoment(ref_el, entity, 2, qdegree))
            entity_ids[1][entity].append(cur)

        super().__init__(nodes, ref_el, entity_ids)


class ArnoldWintherNC(finite_element.CiarletElement):
    """The definition of the nonconforming Arnold-Winther element.
    """
    def __init__(self, ref_el, degree=2):
        if ref_el.shape != TRIANGLE:
            raise ValueError(f"{type(self).__name__} only defined on triangles")
        Ps = polynomial_set.ONSymTensorPolynomialSet(ref_el, degree)
        Ls = ArnoldWintherNCDual(ref_el, degree)
        formdegree = ref_el.get_spatial_dimension() - 1
        mapping = "double contravariant piola"
        super().__init__(Ps, Ls, degree, formdegree, mapping=mapping)


class ArnoldWintherDual(dual_set.DualSet):
    def __init__(self, ref_el, degree=3):
        if degree != 3:
            raise ValueError("Arnold-Winther elements are only defined for degree 3.")
        top = ref_el.get_topology()
        sd = ref_el.get_spatial_dimension()
        shp = (sd, sd)
        entity_ids = {dim: {entity: [] for entity in sorted(top[dim])} for dim in sorted(top)}
        nodes = []

        # vertex dofs
        for v in sorted(top[0]):
            cur = len(nodes)
            pt, = ref_el.make_points(0, v, degree)
            nodes.extend(ComponentPointEvaluation(ref_el, (i, j), shp, pt)
                         for i in range(sd) for j in range(i, sd))
            entity_ids[0][v].extend(range(cur, len(nodes)))

        # edge dofs: bidirectional nn and nt moments against P_{k-2}
        max_order = degree - 2
        qdegree = degree + max_order
        for entity in sorted(top[1]):
            cur = len(nodes)
            for order in range(max_order+1):
                nodes.append(IntegralLegendreNormalNormalMoment(ref_el, entity, order, qdegree))
                nodes.append(IntegralLegendreNormalTangentialMoment(ref_el, entity, order, qdegree))
            entity_ids[1][entity].extend(range(cur, len(nodes)))

        # internal dofs: moments of unique components against P_{k-3}
        n = list(map(ref_el.compute_scaled_normal, sorted(top[sd-1])))
        Q = create_quadrature(ref_el, 2*(degree-1))
        P = polynomial_set.ONPolynomialSet(ref_el, degree-3, scale="L2 piola")
        phis = P.tabulate(Q.get_points())[(0,)*sd]
        nodes.extend(TensorBidirectionalIntegralMoment(ref_el, n[i+1], n[j+1], Q, phi)
                     for phi in phis for i in range(sd) for j in range(i, sd))

        # constraint dofs: moments of divergence against P_{k-1} \ P_{k-2}
        P = polynomial_set.ONPolynomialSet(ref_el, degree-1, shape=(sd,))
        dimPkm1 = P.expansion_set.get_num_members(degree-1)
        dimPkm2 = P.expansion_set.get_num_members(degree-2)
        PH = P.take([i + j * dimPkm1 for j in range(sd) for i in range(dimPkm2, dimPkm1)])
        phis = PH.tabulate(Q.get_points())[(0,)*sd]
        nodes.extend(IntegralMomentOfTensorDivergence(ref_el, Q, phi) for phi in phis)

        entity_ids[2][0].extend(range(cur, len(nodes)))
        super().__init__(nodes, ref_el, entity_ids)


class ArnoldWinther(finite_element.CiarletElement):
    """The definition of the conforming Arnold-Winther element.
    """
    def __init__(self, ref_el, degree=3):
        if ref_el.shape != TRIANGLE:
            raise ValueError(f"{type(self).__name__} only defined on triangles")
        Ps = polynomial_set.ONSymTensorPolynomialSet(ref_el, degree)
        Ls = ArnoldWintherDual(ref_el, degree)
        formdegree = ref_el.get_spatial_dimension() - 1
        mapping = "double contravariant piola"
        super().__init__(Ps, Ls, degree, formdegree, mapping=mapping)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 Miklós Homolya
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

import math
import numpy

from FIAT.finite_element import FiniteElement
from FIAT.dual_set import DualSet
from FIAT.polynomial_set import mis
from FIAT.pointwise_dual import compute_pointwise_dual
from FIAT.reference_element import make_lattice


class BernsteinDualSet(DualSet):
    """The dual basis for Bernstein elements."""

    def __init__(self, ref_el, degree):
        # Initialise data structures
        topology = ref_el.get_topology()
        entity_ids = {dim: {entity_i: []
                            for entity_i in entities}
                      for dim, entities in topology.items()}

        # Calculate inverse topology
        inverse_topology = {vertices: (dim, entity_i)
                            for dim, entities in topology.items()
                            for entity_i, vertices in entities.items()}

        # Generate triangular barycentric indices
        dim = ref_el.get_spatial_dimension()
        kss = mis(dim + 1, degree)

        # Fill data structures
        nodes = []
        for i, ks in enumerate(kss):
            vertices, = numpy.nonzero(ks)
            entity_dim, entity_i = inverse_topology[tuple(vertices)]
            entity_ids[entity_dim][entity_i].append(i)

            # Leave nodes unimplemented for now
            nodes.append(None)

        super().__init__(nodes, ref_el, entity_ids)


class Bernstein(FiniteElement):
    """A finite element with Bernstein polynomials as basis functions."""

    def __init__(self, ref_el, degree):
        dual = BernsteinDualSet(ref_el, degree)
        k = 0  # 0-form
        super().__init__(ref_el, dual, degree, k)
        pts = make_lattice(ref_el.vertices, degree, variant="gll")
        newdual = compute_pointwise_dual(self, pts)
        self.dual = newdual

    def degree(self):
        """The degree of the polynomial space."""
        return self.get_order()

    def value_shape(self):
        """The value shape of the finite element functions."""
        return ()

    def tabulate(self, order, points, entity=None):
        """Return tabulated values of derivatives up to given order of
        basis functions at given points.

        :arg order: The maximum order of derivative.
        :arg points: An iterable of points.
        :arg entity: Optional (dimension, entity number) pair
                     indicating which topological entity of the
                     reference element to tabulate on.  If ``None``,
                     default cell-wise tabulation is performed.
        """
        # Transform points to reference cell coordinates
        ref_el = self.get_reference_element()
        dim = ref_el.get_spatial_dimension()
        if entity is None:
            entity = (dim, 0)

        entity_dim, entity_id = entity
        entity_transform = ref_el.get_entity_transform(entity_dim, entity_id)

        points = numpy.asarray(points)
        cell_points = entity_transform(points)

        # Construct Cartesian to Barycentric coordinate mapping
        vs = numpy.asarray(ref_el.get_vertices())
        B2R = numpy.vstack([vs.T, numpy.ones(len(vs))])
        R2B = numpy.linalg.inv(B2R)

        B = numpy.concatenate([cell_points, numpy.ones((*cell_points.shape[:-1], 1))
                               ], axis=-1).dot(R2B.T)

        # Evaluate everything
        deg = self.degree()
        raw_result = {(alpha, i): vec
                      for i, ks in enumerate(mis(dim + 1, deg))
                      for o in range(order + 1)
                      for alpha, vec in bernstein_Dx(B, ks, o, R2B).items()}

        # Rearrange result
        space_dim = self.space_dimension()
        dtype = numpy.array(list(raw_result.values())).dtype
        result = {alpha: numpy.zeros((space_dim, *points.shape[:-1]), dtype=dtype)
                  for o in range(order + 1)
                  for alpha in mis(dim, o)}
        for (alpha, i), vec in raw_result.items():
            result[alpha][i] = vec
        return result


def bernstein_db(points, ks, alpha=None):
    """Evaluates Bernstein polynomials or its derivative at barycentric
    points.

    :arg points: array of points in barycentric coordinates
    :arg ks: exponents defining the Bernstein polynomial
    :arg alpha: derivative tuple

    :returns: array of Bernstein polynomial values at given points.
    """
    points = numpy.asarray(points)
    ks = numpy.array(tuple(ks))

    *shp, d_1 = points.shape
    assert d_1 == len(ks)

    if alpha is None:
        alpha = numpy.zeros(d_1)
    else:
        alpha = numpy.array(tuple(alpha))
        assert d_1 == len(alpha)

    ls = ks - alpha
    if any(k < 0 for k in ls):
        return numpy.zeros(len(points))
    elif all(k == 0 for k in ls):
        return numpy.ones(len(points))
    else:
        # Calculate coefficient
        coeff = math.factorial(ks.sum())
        for k in ls:
            coeff //= math.factorial(k)
        return coeff * numpy.prod(points**ls, axis=-1)


def bernstein_Dx(points, ks, order, R2B):
    """Evaluates Bernstein polynomials or its derivatives according to
    reference coordinates.

    :arg points: array of points in BARYCENTRIC COORDINATES
    :arg ks: exponents defining the Bernstein polynomial
    :arg alpha: derivative order (returns all derivatives of this
                specified order)
    :arg R2B: linear mapping from reference to barycentric coordinates

    :returns: dictionary mapping from derivative tuples to arrays of
              Bernstein polynomial values at given points.
    """
    points = numpy.asarray(points)
    ks = tuple(ks)

    *shp, d_1 = points.shape
    assert d_1 == len(ks)

    # Collect derivatives according to barycentric coordinates
    Db_map = {alpha: bernstein_db(points, ks, alpha)
              for alpha in mis(d_1, order)}

    # Arrange derivative tensor (barycentric coordinates)
    dtype = numpy.array(list(Db_map.values())).dtype
    Db_shape = (d_1,) * order
    Db_tensor = numpy.empty(Db_shape + tuple(shp), dtype=dtype)
    for ds in numpy.ndindex(Db_shape):
        alpha = tuple(map(ds.count, range(d_1)))
        Db_tensor[ds] = Db_map[alpha]

    # Coordinate transformation: barycentric -> reference
    result = {}
    for alpha in mis(d_1 - 1, order):
        values = Db_tensor
        for d, k in enumerate(alpha):
            for _ in range(k):
                values = R2B[:, d].dot(values)
        result[alpha] = values
    return result



import FIAT
import numpy

from gem import ListTensor, partial_indexed, Literal, Power

from finat.fiat_elements import ScalarFiatElement
from finat.physically_mapped import Citations, identity, PhysicallyMappedElement


class Morley(PhysicallyMappedElement, ScalarFiatElement):
    def __init__(self, cell, degree=2):
        if Citations is not None:
            Citations().register("Morley1971")
            Citations().register("MingXu2006")
        super().__init__(FIAT.Morley(cell, degree=degree))

    def basis_transformation(self, coordinate_mapping):
        sd = self.cell.get_spatial_dimension()
        top = self.cell.get_topology()
        # Jacobians at barycenter
        bary, = self.cell.make_points(sd, 0, sd+1)
        J = coordinate_mapping.jacobian_at(bary)
        detJ = coordinate_mapping.detJ_at(bary)
        V = identity(self.space_dimension())

        offset = len(top[sd-2])
        if sd == 2:
            pel = coordinate_mapping.physical_edge_lengths()
            pts = coordinate_mapping.physical_tangents()
            pns = coordinate_mapping.physical_normals()
            for e in top[sd-1]:
                s = offset + e
                t = partial_indexed(pts, (e,))
                n = partial_indexed(pns, (e,))
                nhat = self.cell.compute_normal(e)
                Jn = J @ Literal(nhat)
                Bnn = Jn @ n
                Bnt = Jn @ t
                V[s, s] = Bnn
                v = list(top[sd-1][e])
                V[s, v] = Bnt / pel[e]
                V[s, v[0]] *= -1

        else:
            adjugate = lambda A: ListTensor([[A[1, 1], -1*A[1, 0]], [-1*A[0, 1], A[0, 0]]])
            edges = self.cell.get_connectivity()[(sd-1, sd-2)]
            for face in top[sd-1]:
                s = offset + face
                thats = self.cell.compute_tangents(sd-1, face)
                nhat = numpy.cross(*thats)
                ahat = numpy.linalg.norm(nhat)
                nhat /= numpy.dot(nhat, nhat)

                Jn = J @ Literal(nhat)
                Jt = J @ Literal(thats.T)
                Gnt = Jn.T @ Jt
                Gtt = Jt.T @ Jt
                detG = Gtt[0, 0]*Gtt[1, 1] - Gtt[0, 1]*Gtt[1, 0]
                area = Power(detG, Literal(0.5))

                Bnn = detJ / area
                Bnt = Gnt @ adjugate(Gtt) / detG
                Bnn *= ahat
                Bnt *= ahat
                V[s, s] = Bnn
                V[s, list(edges[face])] = (-1*(Bnt[0] + Bnt[1]), Bnt[0], Bnt[1])

        # diagonal post-scaling to patch up conditioning
        h = coordinate_mapping.cell_size()
        for face in top[sd-1]:
            s = offset + face
            verts = top[sd-1][face]
            havg = sum(h[v] for v in verts) / len(verts)
            V[:, s] *= 1/havg

        return ListTensor(V.T)



import pytest
import numpy

from FIAT import JohnsonMercier, Nedelec
from FIAT.reference_element import ufc_simplex
from FIAT.quadrature_schemes import create_quadrature


@pytest.fixture(params=("T-ref", "T-phys", "S-ref", "S-phys"))
def cell(request):
    cell, deform = request.param.split("-")
    dim = {"T": 2, "S": 3}[cell]
    K = ufc_simplex(dim)
    if deform == "phys":
        if dim == 2:
            K.vertices = ((0.0, 0.0), (2.0, 0.1), (0.0, 1.0))
        else:
            K.vertices = ((0, 0, 0), (1., 0.1, -0.37),
                          (0.01, 0.987, -.23), (-0.1, -0.2, 1.38))
    return K


def test_johnson_mercier_divergence_rigid_body_motions(cell):
    # test that the divergence of interior JM basis functions is orthogonal to
    # the rigid-body motions
    degree = 1
    variant = None
    sd = cell.get_spatial_dimension()
    JM = JohnsonMercier(cell, degree, variant=variant)

    ref_complex = JM.get_reference_complex()
    Q = create_quadrature(ref_complex, 2*(degree)-1)
    qpts, qwts = Q.get_points(), Q.get_weights()

    tab = JM.tabulate(1, qpts)
    div = sum(tab[alpha][:, alpha.index(1), :, :] for alpha in tab if sum(alpha) == 1)

    # construct rigid body motions
    N1 = Nedelec(cell, 1)
    N1_at_qpts = N1.tabulate(0, qpts)
    rbms = N1_at_qpts[(0,)*sd]
    ells = rbms * qwts[None, None, :]

    edofs = JM.entity_dofs()
    idofs = edofs[sd][0]
    L = numpy.tensordot(div, ells, axes=((1, 2), (1, 2)))
    assert numpy.allclose(L[idofs], 0)

    if variant == "divergence":
        edofs = JM.entity_dofs()
        cdofs = []
        for entity in edofs[sd-1]:
            cdofs.extend(edofs[sd-1][entity][:sd])
        D = L[cdofs]
        M = numpy.tensordot(rbms, ells, axes=((1, 2), (1, 2)))
        X = numpy.linalg.solve(M, D.T)
        assert numpy.allclose(numpy.tensordot(X, rbms, axes=(0, 0)), div[cdofs])



import FIAT

import gem

from finat.fiat_elements import ScalarFiatElement, Lagrange, DiscontinuousLagrange
from finat.point_set import GaussLobattoLegendrePointSet, GaussLegendrePointSet


class GaussLobattoLegendre(Lagrange):
    """1D continuous element with nodes at the Gauss-Lobatto points."""

    def __init__(self, cell, degree):
        fiat_element = FIAT.GaussLobattoLegendre(cell, degree)
        super(Lagrange, self).__init__(fiat_element)

    def basis_evaluation(self, order, ps, entity=None, coordinate_mapping=None):
        '''Return code for evaluating the element at known points on the
        reference element.

        :param order: return derivatives up to this order.
        :param ps: the point set.
        :param entity: the cell entity on which to tabulate.
        '''

        result = super().basis_evaluation(order, ps, entity)
        cell_dimension = self.cell.get_dimension()
        if entity is None or entity == (cell_dimension, 0):  # on cell interior
            space_dim = self.space_dimension()
            if isinstance(ps, GaussLobattoLegendrePointSet) and len(ps.points) == space_dim:
                # Bingo: evaluation points match node locations!
                spatial_dim = self.cell.get_spatial_dimension()
                q, = ps.indices
                r, = self.get_indices()
                result[(0,) * spatial_dim] = gem.ComponentTensor(gem.Delta(q, r), (r,))
        return result


class GaussLegendre(DiscontinuousLagrange):
    """1D discontinuous element with nodes at the Gauss-Legendre points."""

    def __init__(self, cell, degree):
        fiat_element = FIAT.GaussLegendre(cell, degree)
        super(DiscontinuousLagrange, self).__init__(fiat_element)

    def basis_evaluation(self, order, ps, entity=None, coordinate_mapping=None):
        '''Return code for evaluating the element at known points on the
        reference element.

        :param order: return derivatives up to this order.
        :param ps: the point set.
        :param entity: the cell entity on which to tabulate.
        '''

        result = super().basis_evaluation(order, ps, entity)
        cell_dimension = self.cell.get_dimension()
        if entity is None or entity == (cell_dimension, 0):  # on cell interior
            space_dim = self.space_dimension()
            if isinstance(ps, GaussLegendrePointSet) and len(ps.points) == space_dim:
                # Bingo: evaluation points match node locations!
                spatial_dim = self.cell.get_spatial_dimension()
                q, = ps.indices
                r, = self.get_indices()
                result[(0,) * spatial_dim] = gem.ComponentTensor(gem.Delta(q, r), (r,))
        return result


class Legendre(ScalarFiatElement):
    """DG element with Legendre polynomials."""

    def __init__(self, cell, degree, variant=None):
        fiat_element = FIAT.Legendre(cell, degree, variant=variant)
        super().__init__(fiat_element)


class IntegratedLegendre(ScalarFiatElement):
    """CG element with integrated Legendre polynomials."""

    def __init__(self, cell, degree, variant=None):
        fiat_element = FIAT.IntegratedLegendre(cell, degree, variant=variant)
        super().__init__(fiat_element)


class FDMLagrange(ScalarFiatElement):
    """1D CG element with FDM shape functions and point evaluation BCs."""

    def __init__(self, cell, degree):
        fiat_element = FIAT.FDMLagrange(cell, degree)
        super().__init__(fiat_element)


class FDMDiscontinuousLagrange(ScalarFiatElement):
    """1D DG element with derivatives of FDM shape functions with point evaluation Bcs."""

    def __init__(self, cell, degree):
        fiat_element = FIAT.FDMDiscontinuousLagrange(cell, degree)
        super().__init__(fiat_element)


class FDMQuadrature(ScalarFiatElement):
    """1D CG element with FDM shape functions and orthogonalized vertex modes."""

    def __init__(self, cell, degree):
        fiat_element = FIAT.FDMQuadrature(cell, degree)
        super().__init__(fiat_element)


class FDMBrokenH1(ScalarFiatElement):
    """1D Broken CG element with FDM shape functions."""

    def __init__(self, cell, degree):
        fiat_element = FIAT.FDMBrokenH1(cell, degree)
        super().__init__(fiat_element)


class FDMBrokenL2(ScalarFiatElement):
    """1D DG element with derivatives of FDM shape functions."""

    def __init__(self, cell, degree):
        fiat_element = FIAT.FDMBrokenL2(cell, degree)
        super().__init__(fiat_element)


class FDMHermite(ScalarFiatElement):
    """1D CG element with FDM shape functions, point evaluation BCs and derivative BCs."""

    def __init__(self, cell, degree):
        fiat_element = FIAT.FDMHermite(cell, degree)
        super().__init__(fiat_element)



from sympy import symbols, legendre, Array, diff
import numpy as np
from FIAT.finite_element import FiniteElement
from FIAT.dual_set import make_entity_closure_ids
from FIAT.polynomial_set import mis
from FIAT.reference_element import compute_unflattening_map, flatten_reference_cube

x, y, z = symbols('x y z')
variables = (x, y, z)
leg = legendre


def triangular_number(n):
    return int((n+1)*n/2)


def choose_ijk_total(degree):
    top = 1
    for i in range(1, 2 + degree + 1):
        top = i * top
    bottom = 1
    for i in range(1, degree + 1):
        bottom = i * bottom
    return int(top / (2 * bottom))


class TrimmedSerendipity(FiniteElement):
    def __init__(self, ref_el, degree, mapping):
        if degree < 1:
            raise Exception("Trimmed serendipity elements only valid for k >= 1")

        flat_el = flatten_reference_cube(ref_el)
        dim = flat_el.get_spatial_dimension()
        self.fdim = dim
        if dim != 3:
            if dim != 2:
                raise Exception("Trimmed serendipity elements only valid for dimensions 2 and 3")

        flat_topology = flat_el.get_topology()
        entity_ids = {}
        cur = 0
        for top_dim, entities in flat_topology.items():
            entity_ids[top_dim] = {}
            for entity in entities:
                entity_ids[top_dim][entity] = []

        # 3-d case.
        if dim == 3:
            entity_ids[3] = {}
            for j in sorted(flat_topology[2]):
                entity_ids[2][j] = list(range(cur, cur + triangular_number(degree)))
                cur = cur + triangular_number(degree)
            interior_ids = 0
            for k in range(2, degree):
                interior_ids = interior_ids + 3 * choose_ijk_total(k - 2)
            if (degree > 1):
                interior_tilde_ids = 3
                for k in range(1, degree - 1):
                    interior_tilde_ids = interior_tilde_ids + 3
            if (degree == 4):
                interior_tilde_ids += choose_ijk_total(degree - 2) - (degree - 1) - (degree - 1) + 1
            if (degree > 4):
                interior_tilde_ids += choose_ijk_total(degree - 2) - (degree - 1) - (degree - 1) + 1
            if degree == 1:
                interior_tilde_ids = 0
            entity_ids[3][0] = list(range(cur, cur + interior_ids + interior_tilde_ids))
            cur = cur + interior_ids + interior_tilde_ids
        else:
            for j in sorted(flat_topology[1]):
                entity_ids[1][j] = list(range(cur, cur + degree))
                cur = cur + degree

            if (degree >= 2):
                entity_ids[2][0] = list(range(cur, cur + 2*triangular_number(degree - 2) + degree))

            cur += 2*triangular_number(degree - 2) + degree
        formdegree = dim - 1

        entity_closure_ids = make_entity_closure_ids(flat_el, entity_ids)

        super().__init__(ref_el=ref_el,
                         dual=None,
                         order=degree,
                         formdegree=formdegree,
                         mapping=mapping)

        topology = ref_el.get_topology()
        unflattening_map = compute_unflattening_map(topology)
        unflattened_entity_ids = {}
        unflattened_entity_closure_ids = {}

        for dim, entities in sorted(topology.items()):
            unflattened_entity_ids[dim] = {}
            unflattened_entity_closure_ids[dim] = {}
        for dim, entities in sorted(flat_topology.items()):
            for entity in entities:
                unflat_dim, unflat_entity = unflattening_map[(dim, entity)]
                unflattened_entity_ids[unflat_dim][unflat_entity] = entity_ids[dim][entity]
                unflattened_entity_closure_ids[unflat_dim][unflat_entity] = entity_closure_ids[dim][entity]
        self.entity_ids = unflattened_entity_ids
        self.entity_closure_ids = unflattened_entity_closure_ids
        self._degree = degree
        self.flat_el = flat_el

    def degree(self):
        return self._degree

    def get_nodal_basis(self):
        raise NotImplementedError("get_nodal_basis not implemented for trimmed serendipity")

    def get_dual_set(self):
        raise NotImplementedError("get_dual_set is not implemented for trimmed serendipity")

    def get_coeffs(self):
        raise NotImplementedError("get_coeffs not implemented for trimmed serendipity")

    def tabulate(self, order, points, entity=None):

        if entity is None:
            entity = (self.ref_el.get_dimension(), 0)

        entity_dim, entity_id = entity
        transform = self.ref_el.get_entity_transform(entity_dim, entity_id)
        points = transform(points)

        phivals = {}

        for o in range(order+1):
            alphas = mis(self.fdim, o)
            for alpha in alphas:
                try:
                    polynomials = self.basis[alpha]
                except KeyError:
                    zr = tuple([0] * self.fdim)
                    polynomials = diff(self.basis[zr], *zip(variables, alpha))
                    self.basis[alpha] = polynomials
                T = np.zeros((len(polynomials[:, 0]), self.fdim, len(points)))
                for i in range(len(points)):
                    subs = {v: points[i][k] for k, v in enumerate(variables[:self.fdim])}
                    for ell in range(self.fdim):
                        for j, f in enumerate(polynomials[:, ell]):
                            T[j, ell, i] = f.evalf(subs=subs)
                phivals[alpha] = T

        return phivals

    def entity_dofs(self):
        """Return the map of topological entities to degrees of
        freedom for the finite element."""
        return self.entity_ids

    def entity_closure_dofs(self):
        """Return the map of topological entities to degrees of
        freedom on the closure of those entities for the finite element."""
        return self.entity_closure_ids

    def value_shape(self):
        return (self.fdim,)

    def dmats(self):
        raise NotImplementedError

    def get_num_members(self, arg):
        raise NotImplementedError

    def space_dimension(self):
        return int(len(self.basis[tuple([0] * self.fdim)])/self.fdim)


class TrimmedSerendipityDiv(TrimmedSerendipity):
    def __init__(self, ref_el, degree):
        if degree < 1:
            raise Exception("Trimmed serendipity face elements only valid for k >= 1")

        flat_el = flatten_reference_cube(ref_el)
        dim = flat_el.get_spatial_dimension()
        if dim != 2:
            if dim != 3:
                raise Exception("Trimmed serendipity face elements only valid for dimensions 2 and 3")

        verts = flat_el.get_vertices()
        dx = ((verts[-1][0] - x)/(verts[-1][0] - verts[0][0]), (x - verts[0][0])/(verts[-1][0] - verts[0][0]))
        dy = ((verts[-1][1] - y)/(verts[-1][1] - verts[0][1]), (y - verts[0][1])/(verts[-1][1] - verts[0][1]))
        x_mid = 2*x-(verts[-1][0] + verts[0][0])
        y_mid = 2*y-(verts[-1][1] + verts[0][1])
        try:
            dz = ((verts[-1][2] - z)/(verts[-1][2] - verts[0][2]), (z - verts[0][2])/(verts[-1][2] - verts[0][2]))
            z_mid = 2*z-(verts[-1][2] + verts[0][2])
        except IndexError:
            dz = None
            z_mid = None

        if dim == 3:
            FL = f_lambda_2_3d(degree, dx, dy, dz, x_mid, y_mid, z_mid)
            if (degree > 1):
                IL = I_lambda_2_3d(degree, dx, dy, dz, x_mid, y_mid, z_mid)
            else:
                IL = ()
            Sminus_list = FL + IL
            self.basis = {(0, 0, 0): Array(Sminus_list)}
            super().__init__(ref_el=ref_el, degree=degree, mapping="contravariant piola")

        else:
            # Put all 2 dimensional stuff here.
            if degree < 1:
                raise Exception("Trimmed serendipity face elements only valid for k >= 1")

            EL = e_lambda_1_2d_part_one(degree, dx, dy, x_mid, y_mid)
            if degree >= 2:
                FL = trimmed_f_lambda_2d(degree, dx, dy, x_mid, y_mid)
            else:
                FL = ()
            Sminus_list = EL + FL
            Sminus_list = [[-a[1], a[0]] for a in Sminus_list]
            self.basis = {(0, 0): Array(Sminus_list)}
            super().__init__(ref_el=ref_el, degree=degree, mapping="contravariant piola")


def f_lambda_2_3d(degree, dx, dy, dz, x_mid, y_mid, z_mid):

    FL = tuple([(-leg(j, y_mid) * leg(k, z_mid) * a, 0, 0)
                for a in dx for k in range(0, degree) for j in range(0, degree - k)] +
               [(0, leg(j, x_mid) * leg(k, z_mid) * b, 0)
                for b in dy for k in range(0, degree) for j in range(0, degree - k)] +
               [(0, 0, -leg(j, x_mid) * leg(k, y_mid) * c)
                for c in dz for k in range(0, degree) for j in range(0, degree - k)])
    return FL


def I_lambda_2_3d_pieces(current_deg, dx, dy, dz, x_mid, y_mid, z_mid):
    assert current_deg > 1, 'invalid for i = 1'
    ILpiece = tuple([])
    for j in range(0, current_deg - 1):
        for k in range(0, current_deg - 1 - j):
            ILpiece += tuple([(0, 0, -leg(j, x_mid) * leg(k, y_mid) * leg(current_deg - 2 - j - k, z_mid) *
                               dz[0] * dz[1])] +
                             [(0, -leg(j, x_mid) * leg(k, y_mid) * leg(current_deg - 2 - j - k, z_mid) *
                               dy[0] * dy[1], 0)] +
                             [(-leg(j, x_mid) * leg(k, y_mid) * leg(current_deg - 2 - j - k, z_mid) * dx[0] *
                               dx[1], 0, 0)])
    return ILpiece


def I_lambda_2_3d_tilde(degree, dx, dy, dz, x_mid, y_mid, z_mid):
    assert degree > 1, 'invalid for i = 1'
    IL_tilde = tuple([(0, 0, leg(degree - 2, z_mid) * dz[0] * dz[1])] +
                     [(0, leg(degree - 2, y_mid) * dy[0] * dy[1], 0)] +
                     [(leg(degree - 2, x_mid) * dx[0] * dx[1], 0, 0)])
    IL_tilde += tuple([(leg(degree - j - 2, x_mid) * leg(j, y_mid) * dx[0] * dx[1], leg(degree - j - 1, x_mid) *
                        leg(j - 1, y_mid) * dy[0] * dy[1], 0) for j in range(1, degree - 1)] +
                      [(leg(degree - j - 2, x_mid) * leg(j, z_mid) * dx[0] * dx[1], 0, leg(degree - j - 1, x_mid) *
                        leg(j - 1, z_mid) * dz[0] * dz[1]) for j in range(1, degree - 1)] +
                      [(0, leg(degree - j - 2, y_mid) * leg(j, z_mid) * dy[0] * dy[1], leg(degree - j - 1, y_mid) *
                        leg(j - 1, z_mid) * dz[0] * dz[1]) for j in range(1, degree - 1)])
    for k in range(1, degree - 2):
        for l in range(1, degree - 1 - k):
            j = degree - 2 - k - l
            IL_tilde += tuple([(-leg(j, x_mid) * leg(k, y_mid) * leg(l, z_mid) * dx[0] * dx[1],
                                leg(j + 1, x_mid) * leg(k - 1, y_mid) * leg(l, z_mid) * dy[0] * dy[1],
                                -leg(j + 1, x_mid) * leg(k, y_mid) * leg(l - 1, z_mid) * dz[0] * dz[1])])
    return IL_tilde


def I_lambda_2_3d(degree, dx, dy, dz, x_mid, y_mid, z_mid):
    IL = tuple([])
    for j in range(2, degree):
        IL += I_lambda_2_3d_pieces(j, dx, dy, dz, x_mid, y_mid, z_mid)
    IL += I_lambda_2_3d_tilde(degree, dx, dy, dz, x_mid, y_mid, z_mid)
    return IL


# Everything for 2-d should work already.
def e_lambda_1_2d_part_one(deg, dx, dy, x_mid, y_mid):
    EL = tuple(
        [(0, -leg(j, y_mid) * dx[0]) for j in range(deg)] +
        [(0, -leg(j, y_mid) * dx[1]) for j in range(deg)] +
        [(-leg(j, x_mid)*dy[0], 0) for j in range(deg)] +
        [(-leg(j, x_mid)*dy[1], 0) for j in range(deg)])

    return EL


def e_lambda_tilde_1_2d_part_two(deg, dx, dy, x_mid, y_mid):
    ELTilde = tuple([(-leg(deg, x_mid) * dy[0],
                      -leg(deg-1, x_mid) * dx[0] * dx[1] / (deg+1))] +
                    [(-leg(deg, x_mid) * dy[1],
                      leg(deg-1, x_mid) * dx[0] * dx[1] / (deg+1))] +
                    [(-leg(deg-1, y_mid) * dy[0] * dy[1] / (deg+1),
                      -leg(deg, y_mid) * dx[0])] +
                    [(leg(deg-1, y_mid) * dy[0] * dy[1] / (deg+1),
                      -leg(deg, y_mid) * dx[1])])
    return ELTilde


def e_lambda_1_2d(deg, dx, dy, x_mid, y_mid):
    EL = e_lambda_1_2d_part_one(deg, dx, dy, x_mid, y_mid)
    ELTilde = e_lambda_tilde_1_2d_part_two(deg, dx, dy, x_mid, y_mid)

    result = EL + ELTilde
    return result


def determine_f_lambda_portions_2d(deg):
    if (deg < 2):
        DegsOfIteration = []
    else:
        DegsOfIteration = []
        for i in range(2, deg):
            DegsOfIteration += [i]

    return DegsOfIteration


def f_lambda_1_2d_pieces(current_deg, dx, dy, x_mid, y_mid):
    if (current_deg == 2):
        FLpiece = [(leg(0, x_mid) * leg(0, y_mid) * dy[0] * dy[1], 0)]
        FLpiece += [(0, leg(0, x_mid) * leg(0, y_mid) * dx[0] * dx[1])]
    else:
        target_power = current_deg - 2
        FLpiece = tuple([])
        for j in range(0, target_power + 1):
            k = target_power - j
            FLpiece += tuple([(leg(j, x_mid) * leg(k, y_mid) * dy[0] * dy[1], 0)])
            FLpiece += tuple([(0, leg(j, x_mid) * leg(k, y_mid) * dx[0] * dx[1])])
    return FLpiece


def f_lambda_1_2d_trim(deg, dx, dy, x_mid, y_mid):
    DegsOfIteration = determine_f_lambda_portions_2d(deg)
    FL = []
    for i in DegsOfIteration:
        FL += f_lambda_1_2d_pieces(i, dx, dy, x_mid, y_mid)
    return tuple(FL)


def f_lambda_1_2d_tilde(deg, dx, dy, x_mid, y_mid):
    FLTilde = tuple([])
    FLTilde += tuple([(leg(deg - 2, y_mid)*dy[0]*dy[1], 0)])
    FLTilde += tuple([(0, leg(deg - 2, x_mid)*dx[0]*dx[1])])
    for k in range(1, deg - 1):
        FLTilde += tuple([(leg(k, x_mid) * leg(deg - k - 2, y_mid) * dy[0] * dy[1], -leg(k - 1, x_mid) * leg(deg - k - 1, y_mid) * dx[0] * dx[1])])

    return tuple(FLTilde)


def F_lambda_1_2d(deg, dx, dy, x_mid, y_mid):
    FL = []
    for k in range(2, deg):
        for j in range(k-1):
            FL += [(0, leg(j, x_mid)*leg(k-2-j, y_mid)*dx[0]*dx[1])]
            FL += [(leg(k-2-j, x_mid)*leg(j, y_mid)*dy[0]*dy[1], 0)]

    return tuple(FL)


def trimmed_f_lambda_2d(deg, dx, dy, x_mid, y_mid):
    FL = F_lambda_1_2d(deg, dx, dy, x_mid, y_mid)
    FLT = f_lambda_1_2d_tilde(deg, dx, dy, x_mid, y_mid)
    result = FL + FLT
    return result



import re

from FIAT.macro import IsoSplit, AlfeldSplit, WorseyFarinSplit, PowellSabinSplit, PowellSabin12Split

# dicts mapping Lagrange variant names to recursivenodes family names
supported_cg_variants = {
    "spectral": "gll",
    "chebyshev": "lgc",
    "equispaced": "equispaced",
    "gll": "gll"}

supported_dg_variants = {
    "spectral": "gl",
    "chebyshev": "gc",
    "equispaced": "equispaced",
    "equispaced_interior": "equispaced_interior",
    "gll": "gll",
    "gl": "gl"}

supported_splits = {
    "iso": IsoSplit,
    "alfeld": AlfeldSplit,
    "worsey-farin": WorseyFarinSplit,
    "powell-sabin": PowellSabinSplit,
    "powell-sabin(12)": PowellSabin12Split,
}


def check_format_variant(variant, degree):
    splitting, variant = parse_lagrange_variant(variant, integral=True)

    if variant is None:
        variant = "integral"

    match = re.match(r"^integral(?:\((\d+)\))?$", variant)
    if match:
        variant = "integral"
        extra_degree, = match.groups()
        extra_degree = int(extra_degree) if extra_degree is not None else 0
        interpolant_degree = degree + extra_degree
        if interpolant_degree < degree:
            raise ValueError("Warning, quadrature degree should be at least %s" % degree)
    elif variant == "point":
        interpolant_degree = None
    else:
        raise ValueError('Choose either variant="point" or variant="integral"'
                         'or variant="integral(q)"')

    return splitting, variant, interpolant_degree


def parse_lagrange_variant(variant, discontinuous=False, integral=False):
    """Parses variant options for Lagrange elements.

    variant may be a single option or comma-separated pair
    indicating the dof type (integral, equispaced, spectral, etc)
    and the type of splitting to give a macro-element (Alfeld, Powell-Sabin, iso)
    """
    if variant is None:
        variant = "integral" if integral else "equispaced"
    options = variant.replace(" ", "").split(",")
    assert len(options) <= 2

    default = "integral" if integral else "spectral"
    if integral:
        supported_point_variants = {"integral": None, "point": "point"}
    elif discontinuous:
        supported_point_variants = supported_dg_variants
    else:
        supported_point_variants = supported_cg_variants

    # defaults
    splitting = None
    splitting_args = tuple()
    point_variant = supported_point_variants[default]

    for pre_opt in options:
        opt = pre_opt.lower()
        if opt in supported_splits:
            splitting = supported_splits[opt]
        elif opt.startswith("iso"):
            match = re.match(r"^iso(?:\((\d+)\))?$", opt)
            k, = match.groups()
            call_split = IsoSplit
            splitting_args = (int(k),)
        elif opt.startswith("integral"):
            point_variant = opt
        elif opt in supported_point_variants:
            point_variant = supported_point_variants[opt]
        else:
            raise ValueError("Illegal variant option")

    if discontinuous and splitting is not None and point_variant in supported_cg_variants.values():
        raise ValueError("Illegal variant. DG macroelements with DOFs on subcell boundaries are not unisolvent.")
    if len(splitting_args) > 0:
        splitting = lambda T: call_split(T, *splitting_args, point_variant or "gll")
    return splitting, point_variant



"""Pretty-printing GEM expressions."""
from collections import defaultdict
import itertools

from functools import singledispatch

from gem import gem
from gem.node import collect_refcount, post_traversal


class Context(object):
    def __init__(self):
        expr_counter = itertools.count(1)
        self.expr_name = defaultdict(lambda: "${}".format(next(expr_counter)))
        index_counter = itertools.count(1)
        self.index_name = defaultdict(lambda: "i_{}".format(next(index_counter)))
        self.index_names = set()

    def force_expression(self, expr):
        assert isinstance(expr, gem.Node)
        return self.expr_name[expr]

    def expression(self, expr):
        assert isinstance(expr, gem.Node)
        return self.expr_name.get(expr)

    def index(self, index):
        assert isinstance(index, gem.Index)
        if index.name is None:
            name = self.index_name[index]
        elif index.name not in self.index_names:
            name = index.name
            self.index_name[index] = name
        else:
            name_ = index.name
            for i in itertools.count(1):
                name = "{}~{}".format(name_, i)
                if name not in self.index_names:
                    break
        self.index_names.add(name)
        return name


global_context = Context()


def pprint(expression_dags, context=global_context):
    refcount = collect_refcount(expression_dags)

    def force(node):
        if isinstance(node, gem.Variable):
            return False
        if node.shape:
            return True
        if isinstance(node, (gem.Constant, gem.Indexed, gem.FlexiblyIndexed)):
            return False
        return refcount[node] > 1

    for node in post_traversal(expression_dags):
        if force(node):
            context.force_expression(node)

        name = context.expression(node)
        if name is not None:
            print(make_decl(node, name, context), '=', to_str(node, context, top=True))

    for i, root in enumerate(expression_dags):
        name = "#%d" % (i + 1)
        print(make_decl(root, name, context), '=', to_str(root, context))


def make_decl(node, name, ctx):
    result = name
    if node.shape:
        result += '[' + ','.join(map(repr, node.shape)) + ']'
    if node.free_indices:
        result += '{' + ','.join(map(ctx.index, node.free_indices)) + '}'
    return result


def to_str(expr, ctx, prec=None, top=False):
    if not top and ctx.expression(expr):
        result = ctx.expression(expr)
        if expr.free_indices:
            result += '{' + ','.join(map(ctx.index, expr.free_indices)) + '}'
        return result
    else:
        return _to_str(expr, ctx, prec=prec)


@singledispatch
def _to_str(node, ctx, prec):
    raise AssertionError("GEM node expected")


@_to_str.register(gem.Node)
def _to_str_node(node, ctx, prec):
    front_args = [repr(getattr(node, name)) for name in node.__front__]
    back_args = [repr(getattr(node, name)) for name in node.__back__]
    children = [to_str(child, ctx) for child in node.children]
    return "%s(%s)" % (type(node).__name__, ", ".join(front_args + children + back_args))


@_to_str.register(gem.Zero)
def _to_str_zero(node, ctx, prec):
    assert not node.shape
    return "%g" % node.value


@_to_str.register(gem.Literal)
def _to_str_literal(node, ctx, prec):
    if node.shape:
        return repr(node.array.tolist())
    else:
        return "%g" % node.value


@_to_str.register(gem.Variable)
def _to_str_variable(node, ctx, prec):
    return node.name


@_to_str.register(gem.ListTensor)
def _to_str_listtensor(node, ctx, prec):
    def recurse_rank(array):
        if len(array.shape) > 1:
            return '[' + ', '.join(map(recurse_rank, array)) + ']'
        else:
            return '[' + ', '.join(to_str(item, ctx) for item in array) + ']'

    return recurse_rank(node.array)


@_to_str.register(gem.Indexed)
def _to_str_indexed(node, ctx, prec):
    child, = node.children
    result = to_str(child, ctx)
    dimensions = []
    for index in node.multiindex:
        if isinstance(index, gem.Index):
            dimensions.append(ctx.index(index))
        elif isinstance(index, int):
            dimensions.append(str(index))
        else:
            dimensions.append(to_str(index.expression, ctx))
    result += '[' + ','.join(dimensions) + ']'
    return result


@_to_str.register(gem.FlexiblyIndexed)
def _to_str_flexiblyindexed(node, ctx, prec):
    child, = node.children
    result = to_str(child, ctx)
    dimensions = []
    for offset, idxs in node.dim2idxs:
        parts = []
        if offset:
            parts.append(str(offset))
        for index, stride in idxs:
            index_name = ctx.index(index)
            assert stride
            if stride == 1:
                parts.append(index_name)
            else:
                parts.append(index_name + "*" + str(stride))
        if parts:
            dimensions.append(' + '.join(parts))
        else:
            dimensions.append('0')
    if dimensions:
        result += '[' + ','.join(dimensions) + ']'
    return result


@_to_str.register(gem.IndexSum)
def _to_str_indexsum(node, ctx, prec):
    result = 'Sum_{' + ','.join(map(ctx.index, node.multiindex)) + '} ' + to_str(node.children[0], ctx, prec=2)
    if prec is not None and prec > 2:
        result = '({})'.format(result)
    return result


@_to_str.register(gem.ComponentTensor)
def _to_str_componenttensor(node, ctx, prec):
    return to_str(node.children[0], ctx) + '|' + ','.join(ctx.index(i) for i in node.multiindex)


@_to_str.register(gem.Sum)
def _to_str_sum(node, ctx, prec):
    children = [to_str(child, ctx, prec=1) for child in node.children]
    result = " + ".join(children)
    if prec is not None and prec > 1:
        result = "({})".format(result)
    return result


@_to_str.register(gem.Product)
def _to_str_product(node, ctx, prec):
    children = [to_str(child, ctx, prec=3) for child in node.children]
    result = "*".join(children)
    if prec is not None and prec > 3:
        result = "({})".format(result)
    return result


@_to_str.register(gem.MathFunction)
def _to_str_mathfunction(node, ctx, prec):
    child, = node.children
    return node.name + "(" + to_str(child, ctx) + ")"



import FIAT

from finat.fiat_elements import FiatElement


class NodalEnrichedElement(FiatElement):
    """An enriched element with a nodal basis."""

    def __init__(self, elements):
        nodal_enriched = FIAT.NodalEnrichedElement(*(elem.fiat_equivalent
                                                     for elem in elements))
        super().__init__(nodal_enriched)



"""This module defines the UFL finite element classes."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file was originally part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Kristian B. Oelgaard
# Modified by Marie E. Rognes 2010, 2012
# Modified by Massimiliano Leoni, 2016
# Modified by Matthew Scroggs, 2023

from finat.ufl.finiteelementbase import FiniteElementBase
from ufl.sobolevspace import L2

valid_restriction_domains = ("interior", "facet", "face", "edge", "vertex", "reduced")


class RestrictedElement(FiniteElementBase):
    """Represents the restriction of a finite element to a type of cell entity."""

    def __init__(self, element, restriction_domain):
        """Doc."""
        if not isinstance(element, FiniteElementBase):
            raise ValueError("Expecting a finite element instance.")
        if restriction_domain not in valid_restriction_domains:
            raise ValueError(f"Expecting one of the strings: {valid_restriction_domains}")

        FiniteElementBase.__init__(self, "RestrictedElement", element.cell,
                                   element.degree(),
                                   element.quadrature_scheme(),
                                   element.reference_value_shape)

        self._element = element

        self._restriction_domain = restriction_domain

    def __repr__(self):
        """Doc."""
        return f"RestrictedElement({repr(self._element)}, {repr(self._restriction_domain)})"

    @property
    def sobolev_space(self):
        """Doc."""
        if self._restriction_domain == "interior":
            return L2
        else:
            return self._element.sobolev_space

    def is_cellwise_constant(self):
        """Return whether the basis functions of this element is spatially constant over each cell."""
        return self._element.is_cellwise_constant()

    def _is_linear(self):
        """Doc."""
        return self._element._is_linear()

    def sub_element(self):
        """Return the element which is restricted."""
        return self._element

    def mapping(self):
        """Doc."""
        return self._element.mapping()

    def restriction_domain(self):
        """Return the domain onto which the element is restricted."""
        return self._restriction_domain

    def reconstruct(self, **kwargs):
        """Doc."""
        element = self._element.reconstruct(**kwargs)
        return RestrictedElement(element, self._restriction_domain)

    def __str__(self):
        """Format as string for pretty printing."""
        return "<%s>|_{%s}" % (self._element, self._restriction_domain)

    def shortstr(self):
        """Format as string for pretty printing."""
        return "<%s>|_{%s}" % (self._element.shortstr(),
                               self._restriction_domain)

    def symmetry(self):
        r"""Return the symmetry dict, which is a mapping :math:`c_0 \\to c_1`.

        meaning that component :math:`c_0` is represented by component
        :math:`c_1`.  A component is a tuple of one or more ints.
        """
        return self._element.symmetry()

    @property
    def num_sub_elements(self):
        """Return number of sub elements."""
        return self._element.num_sub_elements

    @property
    def sub_elements(self):
        """Return list of sub elements."""
        return self._element.sub_elements

    def num_restricted_sub_elements(self):
        """Return number of restricted sub elements."""
        return 1

    def restricted_sub_elements(self):
        """Return list of restricted sub elements."""
        return (self._element,)

    def variant(self):
        """Doc."""
        return self._element.variant()



# Copyright (C) 2016 Imperial College London and others
#
# This file is part of FIAT.
#
# FIAT is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FIAT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FIAT. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#
# David Ham
# Pablo Brubeck

import pytest
import numpy as np


@pytest.mark.parametrize("degree", range(1, 8))
@pytest.mark.parametrize("dim", (1, 2, 3))
def test_gll_basis_values(dim, degree):
    """Ensure that integrating a simple monomial produces the expected results."""
    from FIAT import GaussLobattoLegendre, create_quadrature, reference_element

    s = reference_element.symmetric_simplex(dim)
    q = create_quadrature(s, 2*degree)
    fe = GaussLobattoLegendre(s, degree)
    tab = fe.tabulate(0, q.pts)[(0,)*dim]

    for test_degree in range(degree + 1):
        v = lambda x: sum(x)**test_degree
        coefs = [n(v) for n in fe.dual.nodes]
        integral = np.dot(coefs, np.dot(tab, q.wts))
        reference = q.integrate(v)
        assert np.allclose(integral, reference, rtol=1e-14)


@pytest.mark.parametrize("dim, degree", [(1, 4), (2, 4), (3, 4)])
def test_edge_dofs(dim, degree):
    """ Ensure edge DOFs are point evaluations at GL points."""
    from FIAT import GaussLobattoLegendre, quadrature, expansions, reference_element

    s = reference_element.symmetric_simplex(dim)
    fe = GaussLobattoLegendre(s, degree)
    ndof = fe.space_dimension()
    assert ndof == expansions.polynomial_dimension(s, degree)

    points = np.zeros((ndof, dim), "d")
    for i, node in enumerate(fe.dual_basis()):
        points[i, :], = node.get_point_dict().keys()

    # Test that edge DOFs are located at the GLL quadrature points
    line = s if dim == 1 else s.construct_subelement(1)
    lr = quadrature.GaussLobattoLegendreQuadratureLineRule(line, degree + 1)
    quadrature_points = lr.get_points()

    entity_dofs = fe.entity_closure_dofs()
    edge_dofs = entity_dofs[1]
    for entity in edge_dofs:
        if len(edge_dofs[entity]) > 0:
            transform = s.get_entity_transform(1, entity)
            assert np.allclose(points[edge_dofs[entity]], transform(quadrature_points))


@pytest.mark.parametrize("dim, degree", [(1, 64), (2, 16), (3, 16)])
def test_interpolation(dim, degree):
    from FIAT import GaussLobattoLegendre, reference_element

    s = reference_element.symmetric_simplex(dim)
    radius = max(np.linalg.norm(s.vertices, axis=-1))
    s.vertices = tuple(map(tuple, np.array(s.vertices) / radius))

    # f = Runge radial function
    A = 25
    r2 = lambda x: np.linalg.norm(x, axis=-1)**2
    f = lambda x: 1/(1 + A*r2(x))

    points = reference_element.make_lattice(s.get_vertices(), 2*degree+1, variant="gl")
    points = np.array(points)
    f_at_pts = f(points)

    k = 1
    errors = []
    degrees = []
    while k <= degree:
        fe = GaussLobattoLegendre(s, k)
        # interpolate f onto FE space: dual evaluation
        coefficients = np.array([v(f) for v in fe.dual_basis()])
        # interpolate FE space onto quadrature points
        tab = fe.tabulate(0, points)[(0,)*dim]
        # compute max error
        errors.append(max(abs(f_at_pts - np.dot(coefficients, tab))))
        degrees.append(k)
        k *= 2

    errors = np.array(errors)
    degrees = np.array(degrees)

    # Test for exponential convergence
    C = np.sqrt(1/A) + np.sqrt(1+1/A)
    assert all(errors < 1.5 * C**-degrees)


@pytest.mark.parametrize("degree", [4, 8, 12, 16])
@pytest.mark.parametrize("dim", [1, 2, 3])
def test_conditioning(dim, degree):
    from FIAT import GaussLobattoLegendre, create_quadrature, reference_element

    s = reference_element.symmetric_simplex(dim)
    rule = create_quadrature(s, 2*degree)
    points = rule.get_points()
    weights = rule.get_weights()

    fe = GaussLobattoLegendre(s, degree)
    phi = fe.tabulate(1, points)
    v = phi[(0,) * dim]
    grads = [phi[alpha] for alpha in phi if sum(alpha) == 1]
    M = np.dot(v, weights[:, None] * v.T)
    K = sum(np.dot(dv, weights[:, None] * dv.T) for dv in grads)

    def cond(A):
        a = np.linalg.eigvalsh(A)
        a = a[abs(a) > 1E-12]
        return max(a) / min(a)

    kappaM = cond(M)
    kappaK = cond(K)
    assert kappaM ** (1/degree) < dim + 1
    assert kappaK ** (1/degree) < dim + 2


if __name__ == '__main__':
    import os
    pytest.main(os.path.abspath(__file__))



import numpy as np
from FIAT import ufc_simplex, MardalTaiWinther, make_quadrature, expansions


def test_dofs():
    line = ufc_simplex(1)
    T = ufc_simplex(2)
    T.vertices = np.asarray([(0.0, 0.0), (1.0, 0.25), (-0.75, 1.1)])
    MTW = MardalTaiWinther(T, 3)

    Qline = make_quadrature(line, 6)

    linebfs = expansions.LineExpansionSet(line)
    linevals = linebfs.tabulate(1, Qline.pts)

    for ed in range(3):
        n = T.compute_scaled_normal(ed)
        wts = np.asarray(Qline.wts)

        vals = MTW.tabulate(0, Qline.pts, (1, ed))[(0, 0)]
        nvals = np.dot(np.transpose(vals, (0, 2, 1)), n)
        normal_moments = np.zeros((9, 2))
        for bf in range(9):
            for k in range(len(Qline.wts)):
                for m in (0, 1):
                    normal_moments[bf, m] += wts[k] * nvals[bf, k] * linevals[m, k]
        right = np.zeros((9, 2))
        right[3*ed, 0] = 1.0
        right[3*ed+2, 1] = 1.0
        assert np.allclose(normal_moments, right)
    for ed in range(3):
        t = T.compute_edge_tangent(ed)
        wts = np.asarray(Qline.wts)

        vals = MTW.tabulate(0, Qline.pts, (1, ed))[(0, 0)]
        tvals = np.dot(np.transpose(vals, (0, 2, 1)), t)
        tangent_moments = np.zeros(9)
        for bf in range(9):
            for k in range(len(Qline.wts)):
                tangent_moments[bf] += wts[k] * tvals[bf, k] * linevals[0, k]
        right = np.zeros(9)
        right[3*ed + 1] = 1.0
        assert np.allclose(tangent_moments, right)



import pytest
import FIAT
import finat
import gem
import numpy as np
from finat.physically_mapped import PhysicalGeometry


class MyMapping(PhysicalGeometry):
    def __init__(self, cell, verts):
        # cell is reference cell, verts is physical vertices
        self.verts = np.asarray(verts)
        self.cell = cell

    def cell_size(self):
        raise NotImplementedError

    def jacobian_at(self, point):
        raise NotImplementedError

    def detJ_at(self, point):
        raise NotImplementedError

    def reference_normals(self):
        raise NotImplementedError

    def physical_normals(self):
        raise NotImplementedError

    def physical_tangents(self):
        raise NotImplementedError

    def physical_edge_lengths(self):
        raise NotImplementedError

    def physical_points(self, ps, entity=None):
        assert entity is None
        prefs = ps.points
        pvs = self.verts
        pps = np.zeros(prefs.shape, dtype=float)
        for i in range(pps.shape[0]):
            pps[i, :] = (pvs[0, :] * (1-prefs[i, 0]) * (1-prefs[i, 1])
                         + pvs[1, :] * (1-prefs[i, 0]) * prefs[i, 1]
                         + pvs[2, :] * prefs[i, 0] * (1-prefs[i, 1])
                         + pvs[3, :] * prefs[i, 0] * prefs[i, 1])
        return gem.Literal(pps)

    def physical_vertices(self):
        return gem.Literal(self.verts)


def get_pts(cell, deg):
    assert cell.shape == FIAT.reference_element.QUADRILATERAL
    L = cell.construct_subelement(1)
    vs = np.asarray(cell.vertices)
    pts = [pt for pt in cell.vertices]
    Lpts = FIAT.reference_element.make_lattice(L.vertices, deg, 1)
    for e in cell.topology[1]:
        Fmap = cell.get_entity_transform(1, e)
        epts = [tuple(Fmap(pt)) for pt in Lpts]
        pts.extend(epts)
    if deg > 3:
        dx0 = (vs[1, :] - vs[0, :]) / (deg-2)
        dx1 = (vs[2, :] - vs[0, :]) / (deg-2)

        internal_nodes = [tuple(vs[0, :] + dx0 * i + dx1 * j)
                          for i in range(1, deg-2)
                          for j in range(1, deg-1-i)]
        pts.extend(internal_nodes)
    return pts


@pytest.mark.parametrize('degree', [1, 2, 3, 4])
def test_kronecker(degree):
    cell = FIAT.ufc_cell("quadrilateral")
    element = finat.DirectSerendipity(cell, degree)
    pts = finat.point_set.PointSet(get_pts(cell, degree))
    vertices = np.asarray(((0.0, 0.0), (1.0, 0.0), (0.1, 1.1), (0.95, 1.01)))
    mapping = MyMapping(cell, vertices)
    z = tuple([0] * cell.get_spatial_dimension())
    vals = element.basis_evaluation(0, pts, coordinate_mapping=mapping)[z]
    numvals = gem.interpreter.evaluate([vals])[0].arr
    assert np.allclose(numvals, np.eye(*numvals.shape))



# Copyright (C) 2016 Imperial College London and others
#
# This file is part of FIAT.
#
# FIAT is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FIAT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FIAT. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#
# Pablo Brubeck

import pytest
import numpy as np


@pytest.mark.parametrize("dim, family, degree", [(dim, f, degree - 1 if f == "DG" else degree)
                                                 for f in ("CG", "DG")
                                                 for dim in range(1, 4)
                                                 for degree in range(1, 7)])
def test_hierarchical_basis_values(dim, family, degree):
    """Ensure that integrating a simple monomial produces the expected results."""
    from FIAT import ufc_simplex, Legendre, IntegratedLegendre, make_quadrature

    s = ufc_simplex(dim)
    q = make_quadrature(s, degree+1)
    if family == "CG":
        fe = IntegratedLegendre(s, degree)
    else:
        fe = Legendre(s, degree)
    tab = fe.tabulate(0, q.pts)[(0,)*dim]

    for test_degree in range(degree + 1):
        v = lambda x: sum(x)**test_degree
        coefs = [n(v) for n in fe.dual.nodes]
        integral = np.dot(coefs, np.dot(tab, q.wts))
        reference = q.integrate(v)
        assert np.allclose(integral, reference, rtol=1e-14)


@pytest.mark.parametrize("family, degree", [(f, degree - 1 if f == "DG" else degree)
                                            for f in ("CG", "DG")
                                            for degree in range(1, 7)])
def test_hierarchical_sparsity(family, degree):
    from FIAT import ufc_simplex, Legendre, IntegratedLegendre, make_quadrature

    s = ufc_simplex(1)
    q = make_quadrature(s, degree+1)
    if family == "CG":
        fe = IntegratedLegendre(s, degree)
        expected = [5 * min(degree, 3) + 3 * max(0, degree-3) - 1, degree + 3]
    else:
        fe = Legendre(s, degree)
        expected = [degree + 1]

    nnz = lambda A: A.size - np.sum(np.isclose(A, 0.0E0, rtol=1E-14))
    moments = lambda v, u: np.dot(np.multiply(v, q.get_weights()), u.T)
    tab = fe.tabulate(len(expected)-1, q.get_points())
    for k, ennz in enumerate(expected):
        A = sum(moments(tab[alpha], tab[alpha]) for alpha in tab if sum(alpha) == k)
        assert nnz(A) == ennz


if __name__ == '__main__':
    import os
    pytest.main(os.path.abspath(__file__))



"""Legacy UFL features."""
# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file was originally part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Kristian B. Oelgaard
# Modified by Marie E. Rognes 2010, 2012
# Modified by Andrew T. T. McRae 2014
# Modified by Lawrence Mitchell 2014
# Modified by Matthew Scroggs, 2023

from finat.ufl.brokenelement import BrokenElement  # noqa: F401
from finat.ufl.enrichedelement import EnrichedElement, NodalEnrichedElement  # noqa: F401
from finat.ufl.finiteelement import FiniteElement  # noqa: F401
from finat.ufl.finiteelementbase import FiniteElementBase  # noqa: F401
from finat.ufl.hdivcurl import HCurlElement, HDivElement, WithMapping, HDiv, HCurl  # noqa: F401
from finat.ufl.mixedelement import MixedElement, TensorElement, VectorElement  # noqa: F401
from finat.ufl.restrictedelement import RestrictedElement  # noqa: F401
from finat.ufl.tensorproductelement import TensorProductElement  # noqa: F401



# Copyright (C) 2010 Anders Logg, 2015 Jan Blechta
#
# This file is part of FIAT.
#
# FIAT is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FIAT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FIAT. If not, see <http://www.gnu.org/licenses/>.
#
# First added:  2010-01-31
# Last changed: 2014-06-30

import pytest
import json
import numpy
import warnings
import os

from FIAT import supported_elements, make_quadrature, ufc_simplex, \
    expansions, reference_element, polynomial_set

# Parameters
tolerance = 1e-8

# Directories
path = os.path.dirname(os.path.abspath(__file__))
ref_path = os.path.join(path, 'fiat-reference-data')
download_script = os.path.join(path, 'scripts', 'download')


class NumpyEncoder(json.JSONEncoder):

    def default(self, obj):
        # If numpy array, convert it to a list and store it in a dict.
        if isinstance(obj, numpy.ndarray):
            data = obj.tolist()
            return dict(__ndarray__=data,
                        dtype=str(obj.dtype),
                        shape=obj.shape)
        # Let the base class default method raise the TypeError
        return json.JSONEncoder(self, obj)


def json_numpy_obj_hook(dct):
    # If dict and have '__ndarray__' as a key, convert it back to ndarray.
    if isinstance(dct, dict) and '__ndarray__' in dct:
        return numpy.asarray(dct['__ndarray__']).reshape(dct['shape'])
    return dct


def load_reference(filename, create_data):
    """Load reference from file. On failure create new file using supplied
    function.
    """
    try:
        # Try loading the reference
        reference = json.load(open(filename, "r"), object_hook=json_numpy_obj_hook)
    except IOError:
        warnings.warn('Reference file "%s" could not be loaded! '
                      'Creating a new reference file!' % filename,
                      RuntimeWarning)

        # Generate data and store for the future
        reference = create_data()
        json.dump(reference, open(filename, "w"), cls=NumpyEncoder)

        # Report failure
        pytest.fail('Comparison to "%s" failed!' % filename)

    return reference


def test_polynomials():
    def create_data():
        ps = polynomial_set.ONPolynomialSet(
            ref_el=reference_element.DefaultTetrahedron(),
            degree=3
        )
        return ps.get_dmats()

    # Try reading reference values
    filename = os.path.join(ref_path, "reference-polynomials.json")
    reference = load_reference(filename, create_data)

    dmats = create_data()

    for dmat, reference_dmat in zip(dmats, reference):
        assert (abs(dmat - reference_dmat) < tolerance).all()


def test_polynomials_1D():
    def create_data():
        ps = polynomial_set.ONPolynomialSet(
            ref_el=reference_element.DefaultLine(),
            degree=3
        )
        return ps.get_dmats()

    # Try reading reference values
    filename = os.path.join(ref_path, "reference-polynomials_1D.json")
    reference = load_reference(filename, create_data)

    dmats = create_data()

    for dmat, reference_dmat in zip(dmats, reference):
        assert (abs(dmat - reference_dmat) < tolerance).all()


def test_expansions():
    def create_data():
        E = reference_element.DefaultTriangle()
        k = 3
        pts = reference_element.make_lattice(E.get_vertices(), k)
        Phis = expansions.ExpansionSet(E)
        phis = Phis.tabulate(k, pts)
        dphis = Phis.tabulate_derivatives(k, pts)
        return phis, dphis

    # Try reading reference values
    filename = os.path.join(ref_path, "reference-expansions.json")
    reference = load_reference(filename, create_data)

    table_phi, table_dphi = create_data()
    reference_table_phi, reference_table_dphi = reference

    # Test raw point data
    diff = numpy.array(table_phi) - numpy.array(reference_table_phi)
    assert (abs(diff) < tolerance).all()

    # Test derivative values
    for entry, reference_entry in zip(table_dphi, reference_table_dphi):
        for point, reference_point in zip(entry, reference_entry):
            value, gradient = point[0], point[1]
            reference_value, reference_gradient = \
                reference_point[0], reference_point[1]
            assert abs(value - reference_value) < tolerance
            diff = numpy.array(gradient) - numpy.array(reference_gradient)
            assert (abs(diff) < tolerance).all()


def test_expansions_jet():
    def create_data():
        latticeK = 2
        n = 1
        order = 2
        E = reference_element.DefaultTetrahedron()
        pts = reference_element.make_lattice(E.get_vertices(), latticeK)
        F = expansions.TetrahedronExpansionSet(E)
        return F.tabulate_jet(n, pts, order)

    filename = os.path.join(ref_path, "reference-expansions-jet.json")
    reference = load_reference(filename, create_data)

    # Test jet data
    data = create_data()
    for datum, reference_datum in zip(data, reference):
        diff = numpy.array(datum) - numpy.array(reference_datum)
        assert (abs(diff) < tolerance).all()


@pytest.fixture(scope="module")
def quadrature_reference_data():
    filename = os.path.join(ref_path, "reference.json")
    try:
        reference = json.load(open(filename, "r"), object_hook=json_numpy_obj_hook)
    except IOError:
        warnings.warn('Reference file "%s" could not be loaded! '
                      'Creating a new reference file!' % filename,
                      RuntimeWarning)
        # No reference data
        reference = {}
    original = reference.copy()
    # Kind of ugly, we rely on destructively modifying the test data
    # we produce here to check if we're missing something (and
    # therefore have to regenerate the reference data).
    yield reference
    if original != reference:
        # Tests introduced some new data, so dump to disk.
        with open(filename, "w") as f:
            json.dump(reference, f, cls=NumpyEncoder)


def quadrature_test_case_name(test_case):
    family, dim, degree = test_case
    return "{}({}, {})".format(family,
                               {1: "interval",
                                2: "triangle",
                                3: "tetrahedron"}[dim],
                               degree)


@pytest.fixture(params=[("Lagrange", 1, 1),
                        ("Lagrange", 1, 2),
                        ("Lagrange", 1, 3),
                        ("Lagrange", 2, 1),
                        ("Lagrange", 2, 2),
                        ("Lagrange", 2, 3),
                        ("Lagrange", 3, 1),
                        ("Lagrange", 3, 2),
                        ("Lagrange", 3, 3),
                        ("Discontinuous Lagrange", 1, 0),
                        ("Discontinuous Lagrange", 1, 1),
                        ("Discontinuous Lagrange", 1, 2),
                        ("Discontinuous Lagrange", 2, 0),
                        ("Discontinuous Lagrange", 2, 1),
                        ("Discontinuous Lagrange", 2, 2),
                        ("Discontinuous Lagrange", 3, 0),
                        ("Discontinuous Lagrange", 3, 1),
                        ("Discontinuous Lagrange", 3, 2),
                        ("Discontinuous Taylor", 1, 0),
                        ("Discontinuous Taylor", 1, 1),
                        ("Discontinuous Taylor", 1, 2),
                        ("Brezzi-Douglas-Marini", 2, 1),
                        ("Brezzi-Douglas-Marini", 2, 2),
                        ("Brezzi-Douglas-Marini", 2, 3),
                        ("Brezzi-Douglas-Marini", 3, 1),
                        ("Brezzi-Douglas-Marini", 3, 2),
                        ("Brezzi-Douglas-Marini", 3, 3),
                        ("Brezzi-Douglas-Fortin-Marini", 2, 2),
                        ("Raviart-Thomas", 2, 1),
                        ("Raviart-Thomas", 2, 2),
                        ("Raviart-Thomas", 2, 3),
                        ("Raviart-Thomas", 3, 1),
                        ("Raviart-Thomas", 3, 2),
                        ("Raviart-Thomas", 3, 3),
                        ("Discontinuous Raviart-Thomas", 2, 1),
                        ("Discontinuous Raviart-Thomas", 2, 2),
                        ("Discontinuous Raviart-Thomas", 2, 3),
                        ("Discontinuous Raviart-Thomas", 3, 1),
                        ("Discontinuous Raviart-Thomas", 3, 2),
                        ("Discontinuous Raviart-Thomas", 3, 3),
                        ("Nedelec 1st kind H(curl)", 2, 1),
                        ("Nedelec 1st kind H(curl)", 2, 2),
                        ("Nedelec 1st kind H(curl)", 2, 3),
                        ("Nedelec 1st kind H(curl)", 3, 1),
                        ("Nedelec 1st kind H(curl)", 3, 2),
                        ("Nedelec 1st kind H(curl)", 3, 3),
                        ("Nedelec 2nd kind H(curl)", 2, 1),
                        ("Nedelec 2nd kind H(curl)", 2, 2),
                        ("Nedelec 2nd kind H(curl)", 2, 3),
                        ("Nedelec 2nd kind H(curl)", 3, 1),
                        ("Nedelec 2nd kind H(curl)", 3, 2),
                        ("Nedelec 2nd kind H(curl)", 3, 3),
                        ("Crouzeix-Raviart", 1, 1),
                        ("Crouzeix-Raviart", 2, 1),
                        ("Crouzeix-Raviart", 3, 1),
                        ("Regge", 2, 0),
                        ("Regge", 2, 1),
                        ("Regge", 2, 2),
                        ("Regge", 3, 0),
                        ("Regge", 3, 1),
                        ("Regge", 3, 2),
                        ("Bubble", 2, 3),
                        ("Bubble", 2, 4),
                        ("Bubble", 2, 5),
                        ("Bubble", 3, 4),
                        ("Bubble", 3, 5),
                        ("Bubble", 3, 6),
                        ("Hellan-Herrmann-Johnson", 2, 0),
                        ("Hellan-Herrmann-Johnson", 2, 1),
                        ("Hellan-Herrmann-Johnson", 2, 2)],
                ids=quadrature_test_case_name)
def quadrature_test_case(request):
    return request.param


def test_quadrature(quadrature_reference_data, quadrature_test_case):
    num_points = 3
    max_derivative = 3

    # Combinations of (family, dim, degree) to test
    def create_data(family, dim, degree):
        '''Create the reference data.
        '''
        kwargs = {}
        if family in {"Crouzeix-Raviart", "Regge", "Hellan-Herrmann-Johnson"}:
            kwargs["variant"] = "point"
        # Get domain and element class
        domain = ufc_simplex(dim)
        ElementClass = supported_elements[family]
        # Create element
        element = ElementClass(domain, degree, **kwargs)
        # Create quadrature points
        quad_rule = make_quadrature(domain, num_points)
        points = quad_rule.get_points()
        # Tabulate at quadrature points
        table = element.tabulate(max_derivative, points)
        return table

    def _perform_test(family, dim, degree, *, reference_table=None):
        '''Test against reference data.
        '''
        table = create_data(family, dim, degree)
        # Check against reference
        for dtuple in reference_table:
            assert eval(dtuple) in table
            assert table[eval(dtuple)].shape == reference_table[dtuple].shape
            diff = table[eval(dtuple)] - reference_table[dtuple]
            assert (abs(diff) < tolerance).all(), \
                "quadrature case %s %s %s failed!" % (family, dim, degree)

    test_case = quadrature_test_case
    try:
        reference = quadrature_reference_data[str(test_case)]
        _perform_test(*test_case, reference_table=reference)
    except KeyError:
        warnings.warn('Reference file does not contain reference "%s"! '
                      'Creating a new reference file!'
                      % (str(test_case)), RuntimeWarning)
        ref = dict([(str(k), v) for k, v in create_data(*test_case).items()])
        quadrature_reference_data[str(test_case)] = ref
        pytest.fail('No reference data for "%s" available' % str(test_case))


if __name__ == '__main__':
    pytest.main(os.path.abspath(__file__))



# Copyright (C) 2016 Imperial College London and others
#
# This file is part of FIAT.
#
# FIAT is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FIAT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FIAT. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#
# Pablo Brubeck

import pytest
import numpy as np


def make_fdm_element(ref_el, family, degree):
    from FIAT import FDMLagrange, FDMDiscontinuousLagrange, FDMBrokenH1, FDMBrokenL2, FDMQuadrature
    if family == "CG":
        return FDMLagrange(ref_el, degree)
    elif family == "DG":
        return FDMDiscontinuousLagrange(ref_el, degree)
    elif family == "BrokenH1":
        return FDMBrokenH1(ref_el, degree)
    elif family == "BrokenL2":
        return FDMBrokenL2(ref_el, degree)
    elif family == "Quadrature":
        return FDMQuadrature(ref_el, degree)


@pytest.mark.parametrize("family, degree", [(f, degree - 1 if f in {"DG", "BrokenL2"} else degree)
                                            for f in ("CG", "DG", "BrokenH1", "BrokenL2", "Quadrature")
                                            for degree in range(1, 7)])
def test_fdm_basis_values(family, degree):
    """Ensure that integrating a simple monomial produces the expected results."""
    from FIAT import ufc_simplex, make_quadrature

    s = ufc_simplex(1)
    q = make_quadrature(s, degree + 1)
    fe = make_fdm_element(s, family, degree)

    tab = fe.tabulate(0, q.pts)[(0,)]

    for test_degree in range(degree + 1):
        coefs = [float(n(lambda x: x[0]**test_degree)) for n in fe.dual.nodes]
        integral = np.dot(coefs, np.dot(tab, q.wts))
        reference = np.dot([x[0]**test_degree
                            for x in q.pts], q.wts)
        assert np.allclose(integral, reference, rtol=1e-14)


@pytest.mark.parametrize("family, degree", [(f, degree - 1 if f in {"DG", "BrokenL2"} else degree)
                                            for f in ("CG", "DG", "BrokenH1", "BrokenL2", "Quadrature")
                                            for degree in range(1, 7)])
def test_fdm_sparsity(family, degree):
    from FIAT import ufc_simplex, make_quadrature

    s = ufc_simplex(1)
    q = make_quadrature(s, degree+1)
    fe = make_fdm_element(s, family, degree)

    if family == "CG":
        expected = [degree + 3, 5 * degree - 1]
    elif family == "DG":
        expected = [degree + 1]
    elif family == "BrokenH1":
        expected = [degree + 1, degree]
    elif family == "BrokenL2":
        expected = [degree + 1]
    elif family == "Quadrature":
        expected = [degree + 1, 3 * degree - 1 - (degree == 1)]

    nnz = lambda A: A.size - np.sum(np.isclose(A, 0.0E0, rtol=1E-14))
    moments = lambda v, u: np.dot(np.multiply(v, q.get_weights()), u.T)
    tab = fe.tabulate(len(expected)-1, q.get_points())
    for k, ennz in enumerate(expected):
        assert nnz(moments(tab[(k, )], tab[(k, )])) == ennz


if __name__ == '__main__':
    import os
    pytest.main(os.path.abspath(__file__))



# Copyright (C) 2008-2012 Robert C. Kirby (Texas Tech University)
#
# Modified 2020 by the same at Baylor University.
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

import numpy

from FIAT import polynomial_set, functional


class DualSet(object):
    def __init__(self, nodes, ref_el, entity_ids, entity_permutations=None):
        nodes, ref_el, entity_ids, entity_permutations = merge_entities(nodes, ref_el, entity_ids, entity_permutations)
        self.nodes = nodes
        self.ref_el = ref_el
        self.entity_ids = entity_ids
        self.entity_permutations = entity_permutations

        # Compute the nodes on the closure of each sub_entity.
        self.entity_closure_ids = {}
        for dim, entities in ref_el.sub_entities.items():
            self.entity_closure_ids[dim] = {}

            for e, sub_entities in entities.items():
                ids = []

                for d, se in sub_entities:
                    ids += self.entity_ids[d][se]
                ids.sort()
                self.entity_closure_ids[d][e] = ids

    def __iter__(self):
        return iter(self.nodes)

    def __len__(self):
        return len(self.nodes)

    def get_nodes(self):
        return self.nodes

    def get_entity_closure_ids(self):
        return self.entity_closure_ids

    def get_entity_ids(self):
        return self.entity_ids

    def get_entity_permutations(self):
        r"""This method returns a nested dictionary that gives, for
        each dimension, for each entity, and for each possible entity
        orientation, the DoF permutation array that maps the entity
        local DoF ordering to the canonical global DoF ordering; see
        :class:`~.Simplex` and :class:`~.UFCQuadrilateral` for how we
        define entity orientations for standard cells.

        The entity permutations `dict` for the degree 4 Lagrange finite
        element on the interval, for instance, is given by:

        .. code-block:: python3

            {0: {0: {0: [0]},
                 1: {0: [0]}},
             1: {0: {0: [0, 1, 2],
                     1: [2, 1, 0]}}}

        Note that there are two entities on dimension ``0`` (vertices),
        each of which has only one possible orientation, while there is
        a single entity on dimension ``1`` (interval), which has two
        possible orientations representing non-reflected and reflected
        intervals.
        """
        if self.entity_permutations is None:
            raise NotImplementedError("entity_permutations not yet implemented for %s" % type(self))
        return self.entity_permutations

    def get_reference_element(self):
        return self.ref_el

    def to_riesz(self, poly_set):
        r"""This method gives the action of the entire dual set
        on each member of the expansion set underlying poly_set.
        Then, applying the linear functionals of the dual set to an
        arbitrary polynomial in poly_set is accomplished by (generalized)
        matrix multiplication.

        For scalar-valued spaces, this produces a matrix
        :\math:`R_{i, j}` such that
        :\math:`\ell_i(f) = \sum_{j} a_j \ell_i(\phi_j)`
        for :\math:`f=\sum_{j} a_j \phi_j`.

        More generally, it will have shape concatenating
        the number of functionals in the dual set, the value shape
        of functions it takes, and the number of members of the
        expansion set.
        """

        # This rather technical code queries the low-level information
        # in pt_dict and deriv_dict
        # for each functional to find out where it evaluates its
        # inputs and/or their derivatives.  Then, it tabulates the
        # expansion set one time for all the function values and
        # another for all of the derivatives.  This circumvents
        # needing to call the to_riesz method of each functional and
        # also limits the number of different calls to tabulate.

        tshape = self.nodes[0].target_shape
        num_nodes = len(self.nodes)
        es = poly_set.get_expansion_set()
        ed = poly_set.get_embedded_degree()
        num_exp = es.get_num_members(poly_set.get_embedded_degree())

        riesz_shape = (num_nodes, *tshape, num_exp)
        mat = numpy.zeros(riesz_shape, "d")

        pts = set()
        dpts = set()
        Qs_to_ells = dict()
        for i, ell in enumerate(self.nodes):
            if len(ell.deriv_dict) > 0:
                dpts.update(ell.deriv_dict.keys())
                continue
            if isinstance(ell, functional.IntegralMoment):
                Q = ell.Q
            else:
                Q = None
                pts.update(ell.pt_dict.keys())
            if Q in Qs_to_ells:
                Qs_to_ells[Q].append(i)
            else:
                Qs_to_ells[Q] = [i]

        Qs_to_pts = {}
        if len(pts) > 0:
            Qs_to_pts[None] = tuple(sorted(pts))
        for Q in Qs_to_ells:
            if Q is not None:
                cur_pts = tuple(map(tuple, Q.pts))
                Qs_to_pts[Q] = cur_pts
                pts.update(cur_pts)

        # Now tabulate the function values
        pts = list(sorted(pts))
        expansion_values = numpy.transpose(es.tabulate(ed, pts))

        for Q in Qs_to_ells:
            ells = Qs_to_ells[Q]
            cur_pts = Qs_to_pts[Q]
            indices = list(map(pts.index, cur_pts))
            wshape = (len(ells), *tshape, len(cur_pts))
            wts = numpy.zeros(wshape, "d")
            if Q is None:
                for i, k in enumerate(ells):
                    ell = self.nodes[k]
                    for pt, wc_list in ell.pt_dict.items():
                        j = cur_pts.index(pt)
                        for (w, c) in wc_list:
                            wts[i][c][j] = w
            else:
                for i, k in enumerate(ells):
                    ell = self.nodes[k]
                    wts[i][ell.comp][:] = ell.f_at_qpts
                qwts = Q.get_weights()
                wts = numpy.multiply(wts, qwts, out=wts)
            mat[ells] += numpy.dot(wts, expansion_values[indices])

        # Tabulate the derivative values that are needed
        max_deriv_order = max(ell.max_deriv_order for ell in self.nodes)
        if max_deriv_order > 0:
            dpts = list(sorted(dpts))
            # It's easiest/most efficient to get derivatives of the
            # expansion set through the polynomial set interface.
            # This is creating a short-lived set to do just this.
            coeffs = numpy.eye(num_exp)
            expansion = polynomial_set.PolynomialSet(self.ref_el, ed, ed, es, coeffs)
            dexpansion_values = expansion.tabulate(dpts, max_deriv_order)

            ells = [k for k, ell in enumerate(self.nodes) if len(ell.deriv_dict) > 0]
            wshape = (len(ells), *tshape, len(dpts))
            dwts = {alpha: numpy.zeros(wshape, "d") for alpha in dexpansion_values if sum(alpha) > 0}
            for i, k in enumerate(ells):
                ell = self.nodes[k]
                for pt, wac_list in ell.deriv_dict.items():
                    j = dpts.index(pt)
                    for (w, alpha, c) in wac_list:
                        dwts[alpha][i][c][j] = w
            for alpha in dwts:
                mat[ells] += numpy.dot(dwts[alpha], dexpansion_values[alpha].T)
        return mat

    def get_indices(self, restriction_domain, take_closure=True):
        """Returns the list of dofs with support on a given restriction domain.

        :arg restriction_domain: can be 'interior', 'vertex', 'edge', 'face' or 'facet'
        :kwarg take_closure: Are we taking the closure of the restriction domain?
        """
        entity_dofs = self.get_entity_ids()
        if restriction_domain == "interior":
            # Return dofs from interior, never taking the closure
            indices = []
            entities = entity_dofs[max(entity_dofs.keys())]
            for (entity, ids) in sorted_by_key(entities):
                indices.extend(ids)
            return indices

        # otherwise return dofs with d <= dim
        if restriction_domain == "vertex":
            dim = 0
        elif restriction_domain == "edge":
            dim = 1
        elif restriction_domain == "face":
            dim = 2
        elif restriction_domain == "facet":
            dim = self.get_reference_element().get_spatial_dimension() - 1
        else:
            raise RuntimeError("Invalid restriction domain")

        is_prodcell = isinstance(max(entity_dofs.keys()), tuple)

        ldim = 0 if take_closure else dim
        indices = []
        for d in range(ldim, dim + 1):
            if is_prodcell:
                for edim in entity_dofs:
                    if sum(edim) == d:
                        entities = entity_dofs[edim]
                        for (entity, ids) in sorted_by_key(entities):
                            indices.extend(ids)
            else:
                entities = entity_dofs[d]
                for (entity, ids) in sorted_by_key(entities):
                    indices.extend(ids)
        return indices


def sorted_by_key(mapping):
    "Sort dict items by key, allowing different key types."
    # Python3 doesn't allow comparing builtins of different type, therefore the typename trick here
    def _key(x):
        return (type(x[0]).__name__, x[0])
    return sorted(mapping.items(), key=_key)


def make_entity_closure_ids(ref_el, entity_ids):
    entity_closure_ids = {}
    for dim, entities in ref_el.sub_entities.items():
        entity_closure_ids[dim] = {}

        for e, sub_entities in entities.items():
            ids = []

            for d, se in sub_entities:
                ids += entity_ids[d][se]
            ids.sort()
            entity_closure_ids[d][e] = ids

    return entity_closure_ids


def lexsort_nodes(ref_el, nodes, entity=None, offset=0):
    """Sort PointEvaluation nodes in lexicographical ordering."""
    if len(nodes) > 1:
        pts = []
        for node in nodes:
            pt, = node.get_point_dict()
            pts.append(pt)
        bary = ref_el.compute_barycentric_coordinates(pts)
        order = list(offset + numpy.lexsort(bary.T))
    else:
        order = list(range(offset, offset + len(nodes)))
    return order


def merge_entities(nodes, ref_el, entity_ids, entity_permutations):
    """Collect DOFs from simplicial complex onto facets of parent cell."""
    parent_cell = ref_el.get_parent()
    if parent_cell is None:
        return nodes, ref_el, entity_ids, entity_permutations
    parent_ids = {}
    parent_permutations = None
    parent_to_children = ref_el.get_parent_to_children()

    if all(isinstance(node, functional.PointEvaluation) for node in nodes):
        # Merge Lagrange dual with lexicographical reordering
        parent_nodes = []
        for dim in sorted(parent_to_children):
            parent_ids[dim] = {}
            for entity in sorted(parent_to_children[dim]):
                cur = len(parent_nodes)
                for child_dim, child_entity in parent_to_children[dim][entity]:
                    parent_nodes.extend(nodes[i] for i in entity_ids[child_dim][child_entity])
                ids = lexsort_nodes(parent_cell, parent_nodes[cur:], entity=(dim, entity), offset=cur)
                parent_ids[dim][entity] = ids
    else:
        # Merge everything else with the same node ordering
        parent_nodes = nodes
        for dim in sorted(parent_to_children):
            parent_ids[dim] = {}
            for entity in sorted(parent_to_children[dim]):
                parent_ids[dim][entity] = []
                for child_dim, child_entity in parent_to_children[dim][entity]:
                    parent_ids[dim][entity].extend(entity_ids[child_dim][child_entity])

    return parent_nodes, parent_cell, parent_ids, parent_permutations



# Copyright (C) 2010 Marie E. Rognes
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Written by Marie E. Rognes <meg@simula.no> based on original
# implementation by Robert C. Kirby.
#
# Last changed: 2010-01-28

import numpy
from FIAT import finite_element, polynomial_set, dual_set, functional
from FIAT.check_format_variant import check_format_variant
from FIAT.quadrature_schemes import create_quadrature
from FIAT.quadrature import FacetQuadratureRule


class CrouzeixRaviartDualSet(dual_set.DualSet):

    def __init__(self, ref_el, degree, variant, interpolant_deg):
        # Get topology dictionary
        sd = ref_el.get_spatial_dimension()
        top = ref_el.get_topology()

        if degree > 1 and sd != 2:
            raise NotImplementedError("High-order Crouzeix-Raviart is only implemented on triangles.")

        # Initialize empty nodes and entity_ids
        entity_ids = {dim: {entity: [] for entity in top[dim]} for dim in top}
        nodes = []

        # Construct nodes and entity_ids
        if variant == "integral":
            for dim in sorted(top):
                if dim == 0 and dim != sd-1:
                    # Skip vertex dofs
                    continue
                facet = ref_el.construct_subelement(dim)
                if dim == 0:
                    Q_facet = create_quadrature(facet, degree + interpolant_deg-1)
                    Phis = numpy.ones((1, len(Q_facet.pts)))
                else:
                    k = degree - 1 if dim == sd-1 else degree - (1+dim)
                    if k < 0:
                        continue
                    Q_facet = create_quadrature(facet, k + interpolant_deg)
                    poly_set = polynomial_set.ONPolynomialSet(facet, k)
                    Phis = poly_set.tabulate(Q_facet.get_points())[(0,) * dim]

                for i in sorted(top[dim]):
                    cur = len(nodes)
                    Q = FacetQuadratureRule(ref_el, dim, i, Q_facet)
                    scale = 1 / Q.jacobian_determinant()
                    phis = scale * Phis
                    nodes.extend(functional.IntegralMoment(ref_el, Q, phi) for phi in phis)
                    entity_ids[dim][i].extend(range(cur, len(nodes)))
        else:
            for dim in sorted(top):
                if dim == 0 and dim != sd-1:
                    # Skip vertex dofs
                    continue
                for i in sorted(top[dim]):
                    cur = len(nodes)
                    if dim == sd-1 and dim != 0:
                        pts = ref_el.make_points(dim, i, degree-1, variant="gl", interior=0)
                    else:
                        pts = ref_el.make_points(dim, i, degree, variant="gll")
                    nodes.extend(functional.PointEvaluation(ref_el, x) for x in pts)
                    entity_ids[dim][i].extend(range(cur, len(nodes)))

        super().__init__(nodes, ref_el, entity_ids)


class CrouzeixRaviart(finite_element.CiarletElement):
    """The Crouzeix-Raviart finite element:

    K:                 Triangle/Tetrahedron
    Polynomial space:  P_k
    Dual basis:        Evaluation at points or integral moments
    """

    def __init__(self, ref_el, degree, variant=None):
        if degree % 2 != 1:
            raise ValueError("Crouzeix-Raviart only defined for odd degree")

        splitting, variant, interpolant_deg = check_format_variant(variant, degree)
        if splitting is not None:
            ref_el = splitting(ref_el)

        poly_set = polynomial_set.ONPolynomialSet(ref_el, degree)
        dual = CrouzeixRaviartDualSet(ref_el, degree, variant, interpolant_deg)
        super().__init__(poly_set, dual, degree)



import FIAT
import gem
import numpy as np
import sympy as sp
from gem.utils import cached_property

from finat.finiteelementbase import FiniteElementBase
from finat.point_set import PointSet
from finat.sympy2gem import sympy2gem

try:
    from firedrake_citations import Citations
    Citations().add("Geevers2018new", """
@article{Geevers2018new,
 title={New higher-order mass-lumped tetrahedral elements for wave propagation modelling},
 author={Geevers, Sjoerd and Mulder, Wim A and van der Vegt, Jaap JW},
 journal={SIAM journal on scientific computing},
 volume={40},
 number={5},
 pages={A2830--A2857},
 year={2018},
 publisher={SIAM},
 doi={https://doi.org/10.1137/18M1175549},
}
""")
    Citations().add("Chin1999higher", """
@article{chin1999higher,
 title={Higher-order triangular and tetrahedral finite elements with mass lumping for solving the wave equation},
 author={Chin-Joe-Kong, MJS and Mulder, Wim A and Van Veldhuizen, M},
 journal={Journal of Engineering Mathematics},
 volume={35},
 number={4},
 pages={405--426},
 year={1999},
 publisher={Springer},
 doi={https://doi.org/10.1023/A:1004420829610},
}
""")
except ImportError:
    Citations = None


class FiatElement(FiniteElementBase):
    """Base class for finite elements for which the tabulation is provided
    by FIAT."""
    def __init__(self, fiat_element):
        super().__init__()
        self._element = fiat_element

    @property
    def cell(self):
        return self._element.get_reference_element()

    @property
    def complex(self):
        return self._element.get_reference_complex()

    @property
    def degree(self):
        # Requires FIAT.CiarletElement
        return self._element.degree()

    @property
    def formdegree(self):
        return self._element.get_formdegree()

    def entity_dofs(self):
        return self._element.entity_dofs()

    def entity_closure_dofs(self):
        return self._element.entity_closure_dofs()

    @property
    def entity_permutations(self):
        return self._element.entity_permutations()

    def space_dimension(self):
        return self._element.space_dimension()

    @property
    def index_shape(self):
        return (self.space_dimension(),)

    @property
    def value_shape(self):
        return self._element.value_shape()

    @property
    def fiat_equivalent(self):
        # Just return the underlying FIAT element
        return self._element

    def basis_evaluation(self, order, ps, entity=None, coordinate_mapping=None):
        '''Return code for evaluating the element at known points on the
        reference element.

        :param order: return derivatives up to this order.
        :param ps: the point set.
        :param entity: the cell entity on which to tabulate.
        '''
        space_dimension = self._element.space_dimension()
        value_size = np.prod(self._element.value_shape(), dtype=int)
        fiat_result = self._element.tabulate(order, ps.points, entity)
        result = {}
        # In almost all cases, we have
        # self.space_dimension() == self._element.space_dimension()
        # But for Bell, FIAT reports 21 basis functions,
        # but FInAT only 18 (because there are actually 18
        # basis functions, and the additional 3 are for
        # dealing with transformations between physical
        # and reference space).
        index_shape = (self._element.space_dimension(),)
        for alpha, fiat_table in fiat_result.items():
            if isinstance(fiat_table, Exception):
                result[alpha] = gem.Failure(self.index_shape + self.value_shape, fiat_table)
                continue

            derivative = sum(alpha)
            shp = (space_dimension, value_size, *ps.points.shape[:-1])
            table_roll = np.moveaxis(fiat_table.reshape(shp), 0, -1)

            exprs = []
            for table in table_roll:
                if derivative == self.degree and not self.complex.is_macrocell():
                    # Make sure numerics satisfies theory
                    exprs.append(gem.Literal(table[0]))
                elif derivative > self.degree:
                    # Make sure numerics satisfies theory
                    assert np.allclose(table, 0.0)
                    exprs.append(gem.Literal(np.zeros(self.index_shape)))
                else:
                    point_indices = ps.indices
                    point_shape = tuple(index.extent for index in point_indices)

                    exprs.append(gem.partial_indexed(
                        gem.Literal(table.reshape(point_shape + index_shape)),
                        point_indices
                    ))
            if self.value_shape:
                # As above, this extent may be different from that
                # advertised by the finat element.
                beta = tuple(gem.Index(extent=i) for i in index_shape)
                assert len(beta) == len(self.get_indices())

                zeta = self.get_value_indices()
                result[alpha] = gem.ComponentTensor(
                    gem.Indexed(
                        gem.ListTensor(np.array(
                            [gem.Indexed(expr, beta) for expr in exprs]
                        ).reshape(self.value_shape)),
                        zeta),
                    beta + zeta
                )
            else:
                expr, = exprs
                result[alpha] = expr
        return result

    def point_evaluation(self, order, refcoords, entity=None, coordinate_mapping=None):
        '''Return code for evaluating the element at an arbitrary points on
        the reference element.

        :param order: return derivatives up to this order.
        :param refcoords: GEM expression representing the coordinates
                          on the reference entity.  Its shape must be
                          a vector with the correct dimension, its
                          free indices are arbitrary.
        :param entity: the cell entity on which to tabulate.
        :param coordinate_mapping: a
           :class:`~.physically_mapped.PhysicalGeometry` object that
           provides physical geometry callbacks (may be None).
        '''
        if entity is None:
            entity = (self.cell.get_dimension(), 0)
        entity_dim, entity_i = entity

        # Spatial dimension of the entity
        esd = self.cell.construct_subelement(entity_dim).get_spatial_dimension()
        assert isinstance(refcoords, gem.Node) and refcoords.shape == (esd,)

        return point_evaluation(self._element, order, refcoords, (entity_dim, entity_i))

    @cached_property
    def _dual_basis(self):
        # Return the numerical part of the dual basis, this split is
        # needed because the dual_basis itself can't produce the same
        # point set over and over in case it is used multiple times
        # (in for example a tensorproductelement).
        fiat_dual_basis = self._element.dual_basis()
        seen = dict()
        allpts = []
        # Find the unique points to evaluate at.
        # We might be able to make this a smaller set by treating each
        # point one by one, but most of the redundancy comes from
        # multiple functionals using the same quadrature rule.
        for dual in fiat_dual_basis:
            if len(dual.deriv_dict) != 0:
                raise NotImplementedError("FIAT dual bases with derivative nodes represented via a ``Functional.deriv_dict`` property do not currently have a FInAT dual basis")
            pts = dual.get_point_dict().keys()
            pts = tuple(sorted(pts))  # need this for determinism
            if pts not in seen:
                # k are indices into Q (see below) for the seen points
                kstart = len(allpts)
                kend = kstart + len(pts)
                seen[pts] = kstart, kend
                allpts.extend(pts)
        # Build Q.
        # Q is a tensor of weights (of total rank R) to contract with a unique
        # vector of points to evaluate at, giving a tensor (of total rank R-1)
        # where the first indices (rows) correspond to a basis functional
        # (node).
        # Q is a DOK Sparse matrix in (row, col, higher,..)=>value pairs (to
        # become a gem.SparseLiteral when implemented).
        # Rows (i) are number of nodes/dual functionals.
        # Columns (k) are unique points to evaluate.
        # Higher indices (*cmp) are tensor indices of the weights when weights
        # are tensor valued.
        Q = {}
        for i, dual in enumerate(fiat_dual_basis):
            point_dict = dual.get_point_dict()
            pts = tuple(sorted(point_dict.keys()))
            kstart, kend = seen[pts]
            for p, k in zip(pts, range(kstart, kend)):
                for weight, cmp in point_dict[p]:
                    Q[(i, k, *cmp)] = weight
        if all(len(set(key)) == 1 and np.isclose(weight, 1) and len(key) == 2
               for key, weight in Q.items()):
            # Identity matrix Q can be expressed symbolically
            extents = tuple(map(max, zip(*Q.keys())))
            js = tuple(gem.Index(extent=e+1) for e in extents)
            assert len(js) == 2
            Q = gem.ComponentTensor(gem.Delta(*js), js)
        else:
            # temporary until sparse literals are implemented in GEM which will
            # automatically convert a dictionary of keys internally.
            # TODO the below is unnecessarily slow and would be sped up
            # significantly by building Q in a COO format rather than DOK (i.e.
            # storing coords and associated data in (nonzeros, entries) shaped
            # numpy arrays) to take advantage of numpy multiindexing
            if len(Q) == 1:
                Qshape = tuple(s + 1 for s in tuple(Q)[0])
            else:
                Qshape = tuple(s + 1 for s in map(max, *Q))
            Qdense = np.zeros(Qshape, dtype=np.float64)
            for idx, value in Q.items():
                Qdense[idx] = value
            Q = gem.Literal(Qdense)
        return Q, np.asarray(allpts)

    @property
    def dual_basis(self):
        # Return Q with x.indices already a free index for the
        # consumer to use
        # expensive numerical extraction is done once per element
        # instance, but the point set must be created every time we
        # build the dual.
        Q, pts = self._dual_basis
        x = PointSet(pts)
        assert len(x.indices) == 1
        assert Q.shape[1] == x.indices[0].extent
        i, *js = gem.indices(len(Q.shape) - 1)
        Q = gem.ComponentTensor(gem.Indexed(Q, (i, *x.indices, *js)), (i, *js))
        return Q, x

    @property
    def mapping(self):
        mappings = set(self._element.mapping())
        if len(mappings) != 1:
            return None
        else:
            result, = mappings
            return result


def point_evaluation(fiat_element, order, refcoords, entity):
    # Coordinates on the reference entity (SymPy)
    esd, = refcoords.shape
    Xi = sp.symbols('X Y Z')[:esd]

    space_dimension = fiat_element.space_dimension()
    value_size = np.prod(fiat_element.value_shape(), dtype=int)
    fiat_result = fiat_element.tabulate(order, [Xi], entity)
    result = {}
    for alpha, fiat_table in fiat_result.items():
        if isinstance(fiat_table, Exception):
            result[alpha] = gem.Failure((space_dimension,) + fiat_element.value_shape(), fiat_table)
            continue

        # Convert SymPy expression to GEM
        mapper = gem.node.Memoizer(sympy2gem)
        mapper.bindings = {s: gem.Indexed(refcoords, (i,))
                           for i, s in enumerate(Xi)}
        gem_table = np.vectorize(mapper)(fiat_table)

        table_roll = gem_table.reshape(space_dimension, value_size).transpose()

        exprs = []
        for table in table_roll:
            exprs.append(gem.ListTensor(table.reshape(space_dimension)))
        if fiat_element.value_shape():
            beta = (gem.Index(extent=space_dimension),)
            zeta = tuple(gem.Index(extent=d)
                         for d in fiat_element.value_shape())
            result[alpha] = gem.ComponentTensor(
                gem.Indexed(
                    gem.ListTensor(np.array(
                        [gem.Indexed(expr, beta) for expr in exprs]
                    ).reshape(fiat_element.value_shape())),
                    zeta),
                beta + zeta
            )
        else:
            expr, = exprs
            result[alpha] = expr
    return result


class Regge(FiatElement):  # symmetric matrix valued
    def __init__(self, cell, degree, variant=None):
        super().__init__(FIAT.Regge(cell, degree, variant=variant))


class HellanHerrmannJohnson(FiatElement):  # symmetric matrix valued
    def __init__(self, cell, degree, variant=None):
        super().__init__(FIAT.HellanHerrmannJohnson(cell, degree, variant=variant))


class GopalakrishnanLedererSchoberlFirstKind(FiatElement):  # traceless matrix valued
    def __init__(self, cell, degree, variant=None):
        super().__init__(FIAT.GopalakrishnanLedererSchoberlFirstKind(cell, degree, variant=variant))


class GopalakrishnanLedererSchoberlSecondKind(FiatElement):  # traceless matrix valued
    def __init__(self, cell, degree, variant=None):
        super().__init__(FIAT.GopalakrishnanLedererSchoberlSecondKind(cell, degree, variant=variant))


class ScalarFiatElement(FiatElement):
    @property
    def value_shape(self):
        return ()


class Bernstein(ScalarFiatElement):
    # TODO: Replace this with a smarter implementation
    def __init__(self, cell, degree):
        super().__init__(FIAT.Bernstein(cell, degree))


class Bubble(ScalarFiatElement):
    def __init__(self, cell, degree, variant=None):
        super().__init__(FIAT.Bubble(cell, degree, variant=variant))


class FacetBubble(ScalarFiatElement):
    def __init__(self, cell, degree, variant=None):
        super().__init__(FIAT.FacetBubble(cell, degree, variant=variant))


class CrouzeixRaviart(ScalarFiatElement):
    def __init__(self, cell, degree, variant=None):
        super().__init__(FIAT.CrouzeixRaviart(cell, degree, variant=variant))


class Lagrange(ScalarFiatElement):
    def __init__(self, cell, degree, variant=None):
        super().__init__(FIAT.Lagrange(cell, degree, variant=variant))


class KongMulderVeldhuizen(ScalarFiatElement):
    def __init__(self, cell, degree):
        super().__init__(FIAT.KongMulderVeldhuizen(cell, degree))
        if Citations is not None:
            Citations().register("Chin1999higher")
            Citations().register("Geevers2018new")


class DiscontinuousLagrange(ScalarFiatElement):
    def __init__(self, cell, degree, variant=None):
        super().__init__(FIAT.DiscontinuousLagrange(cell, degree, variant=variant))


class Histopolation(ScalarFiatElement):
    def __init__(self, cell, degree):
        super().__init__(FIAT.Histopolation(cell, degree))


class Real(DiscontinuousLagrange):
    ...


class Serendipity(ScalarFiatElement):
    def __init__(self, cell, degree):
        super().__init__(FIAT.Serendipity(cell, degree))


class DPC(ScalarFiatElement):
    def __init__(self, cell, degree):
        super().__init__(FIAT.DPC(cell, degree))


class DiscontinuousTaylor(ScalarFiatElement):
    def __init__(self, cell, degree):
        super().__init__(FIAT.DiscontinuousTaylor(cell, degree))


class HDivTrace(ScalarFiatElement):
    def __init__(self, cell, degree, variant=None):
        super().__init__(FIAT.HDivTrace(cell, degree, variant=variant))


class VectorFiatElement(FiatElement):
    @property
    def value_shape(self):
        return (self.cell.get_spatial_dimension(),)


class RaviartThomas(VectorFiatElement):
    def __init__(self, cell, degree, variant=None):
        super().__init__(FIAT.RaviartThomas(cell, degree, variant=variant))


class TrimmedSerendipityFace(VectorFiatElement):
    def __init__(self, cell, degree):
        super().__init__(FIAT.TrimmedSerendipityFace(cell, degree))

    @property
    def entity_permutations(self):
        raise NotImplementedError(f"entity_permutations not yet implemented for {type(self)}")


class TrimmedSerendipityDiv(VectorFiatElement):
    def __init__(self, cell, degree):
        super().__init__(FIAT.TrimmedSerendipityDiv(cell, degree))

    @property
    def entity_permutations(self):
        raise NotImplementedError(f"entity_permutations not yet implemented for {type(self)}")


class TrimmedSerendipityEdge(VectorFiatElement):
    def __init__(self, cell, degree):
        super().__init__(FIAT.TrimmedSerendipityEdge(cell, degree))

    @property
    def entity_permutations(self):
        raise NotImplementedError(f"entity_permutations not yet implemented for {type(self)}")


class TrimmedSerendipityCurl(VectorFiatElement):
    def __init__(self, cell, degree):
        super().__init__(FIAT.TrimmedSerendipityCurl(cell, degree))

    @property
    def entity_permutations(self):
        raise NotImplementedError(f"entity_permutations not yet implemented for {type(self)}")


class BrezziDouglasMarini(VectorFiatElement):
    def __init__(self, cell, degree, variant=None):
        super().__init__(FIAT.BrezziDouglasMarini(cell, degree, variant=variant))


class BrezziDouglasMariniCubeEdge(VectorFiatElement):
    def __init__(self, cell, degree):
        super().__init__(FIAT.BrezziDouglasMariniCubeEdge(cell, degree))

    @property
    def entity_permutations(self):
        raise NotImplementedError(f"entity_permutations not yet implemented for {type(self)}")


class BrezziDouglasMariniCubeFace(VectorFiatElement):
    def __init__(self, cell, degree):
        super().__init__(FIAT.BrezziDouglasMariniCubeFace(cell, degree))

    @property
    def entity_permutations(self):
        raise NotImplementedError(f"entity_permutations not yet implemented for {type(self)}")


class BrezziDouglasFortinMarini(VectorFiatElement):
    def __init__(self, cell, degree):
        super().__init__(FIAT.BrezziDouglasFortinMarini(cell, degree))


class Nedelec(VectorFiatElement):
    def __init__(self, cell, degree, variant=None):
        super().__init__(FIAT.Nedelec(cell, degree, variant=variant))


class NedelecSecondKind(VectorFiatElement):
    def __init__(self, cell, degree, variant=None):
        super().__init__(FIAT.NedelecSecondKind(cell, degree, variant=variant))
